"""
Light cone halo mask functions.

Author: Britton Smith <brittons@origins.colorado.edu>
Affiliation: CASA/University of CO, Boulder
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 Britton Smith.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import copy
import h5py
import numpy as np

from yt.funcs import *
from yt.analysis_modules.halo_profiler.api import \
     HaloProfiler
from yt.convenience import load
from yt.utilities.parallel_tools.parallel_analysis_interface import \
     parallel_objects, \
     parallel_root_only

def _light_cone_halo_mask(lightCone, cube_file=None,
                          mask_file=None, map_file=None,
                          halo_profiler_parameters=None,
                          virial_overdensity=200,
                          njobs=1, dynamic=False):
    "Make a boolean mask to cut clusters out of light cone projections."

    if halo_profiler_parameters is None:
        halo_profiler_parameters = {}

    pixels = int(lightCone.field_of_view_in_arcminutes * 60.0 /
                 lightCone.image_resolution_in_arcseconds)

    # Loop through files in light cone solution and get virial quantities.
    halo_map_storage = {}
    for my_storage, my_slice in \
      parallel_objects(lightCone.light_cone_solution,
                       njobs=njobs, dynamic=dynamic,
                       storage=halo_map_storage):
        halo_list = _get_halo_list(my_slice['filename'],
                                   **halo_profiler_parameters)
        my_storage.result = \
          {'mask': _make_slice_mask(my_slice, halo_list, pixels,
                                    virial_overdensity)}
        if map_file is not None:
            my_storage.result['map'] = \
              _make_slice_halo_map(my_slice, halo_list,
                                   virial_overdensity)

    # Reassemble halo mask and map lists.
    light_cone_mask = []
    halo_map = []
    all_slices = halo_map_storage.keys()
    all_slices.sort()
    for i in all_slices:
        light_cone_mask.append(halo_map_storage[i]['mask'])
        if map_file is not None:
            halo_map.extend(halo_map_storage[i]['map'])
    del halo_map_storage

    # Write out cube of masks from each slice.
    if cube_file is not None:
        _write_halo_mask(cube_file, np.array(light_cone_mask))

    # Write out a text list of all halos in the image.
    if map_file is not None:
        _write_halo_map(map_file, halo_map)

    # Write out final mask.
    if mask_file is not None:
        # Final mask is simply the product of the mask from each slice.
        final_mask = np.ones(shape=(pixels, pixels))
        for mask in light_cone_mask:
            final_mask *= mask
        _write_halo_mask(mask_file, final_mask)

    return light_cone_mask

@parallel_root_only
def _write_halo_mask(filename, halo_mask):
    r"""Write out an hdf5 file with the halo mask that
    can be applied to an image.
    """

    mylog.info("Saving halo mask to %s." % filename)
    output = h5py.File(filename, 'a')
    if 'HaloMask' in output.keys():
        del output['HaloMask']
    output.create_dataset('HaloMask', data=np.array(halo_mask))
    output.close()

@parallel_root_only
def _write_halo_map(filename, halo_map):
    "Write a text list of halos in a light cone image."

    mylog.info("Saving halo map to %s." % filename)
    f = open(filename, 'w')
    f.write("#z       x         y        r_image   r_mpc     m_Msun\n")
    for halo in halo_map:
        f.write("%7.4f %9.6f %9.6f %9.3e %9.3e %9.3e\n" % \
                    (halo['redshift'], halo['x'], halo['y'],
                     halo['radius_image'], halo['radius_mpc'],
                     halo['mass']))
    f.close()

def _get_halo_list(dataset, halo_profiler_kwargs=None,
                   halo_profiler_actions=None, halo_list='all'):
    "Load a list of halos for the dataset."

    if halo_profiler_kwargs is None: halo_profiler_kwargs = {}
    if halo_profiler_actions is None: halo_profiler_actions = []

    hp = HaloProfiler(dataset, **halo_profiler_kwargs)
    for action in halo_profiler_actions:
        if not action.has_key('args'): action['args'] = ()
        if not action.has_key('kwargs'): action['kwargs'] = {}
        action['function'](hp, *action['args'], **action['kwargs'])

    if halo_list == 'all':
        return_list = copy.deepcopy(hp.all_halos)
    elif halo_list == 'filtered':
        return_list = copy.deepcopy(hp.filtered_halos)
    else:
        mylog.error("Keyword, halo_list, must be either 'all' or 'filtered'.")
        return_list = None

    del hp
    return return_list

def _make_slice_mask(slice, halo_list, pixels, virial_overdensity):
    "Make halo mask for one slice in light cone solution."

    # Get shifted, tiled halo list.
    all_halo_x, all_halo_y, \
      all_halo_radius, all_halo_mass = \
      _make_slice_halo_list(slice, halo_list, virial_overdensity)

    # Make boolean mask and cut out halos.
    dx = slice['box_width_fraction'] / pixels
    x = [(q + 0.5) * dx for q in range(pixels)]
    haloMask = np.ones(shape=(pixels, pixels), dtype=bool)

    # Cut out any pixel that has any part at all in the circle.
    for q in range(len(all_halo_radius)):
        dif_xIndex = np.array(int(all_halo_x[q]/dx) -
                              np.array(range(pixels))) != 0
        dif_yIndex = np.array(int(all_halo_y[q]/dx) -
                              np.array(range(pixels))) != 0

        xDistance = (np.abs(x - all_halo_x[q]) -
                     (0.5 * dx)) * dif_xIndex
        yDistance = (np.abs(x - all_halo_y[q]) -
                     (0.5 * dx)) * dif_yIndex

        distance = np.array([np.sqrt(w**2 + xDistance**2)
                             for w in yDistance])
        haloMask *= (distance >= all_halo_radius[q])

    return haloMask

def _make_slice_halo_map(slice, halo_list, virial_overdensity):
    "Make list of halos for one slice in light cone solution."

    # Get units to convert virial radii back to physical units.
    dataset_object = load(slice['filename'])
    Mpc_units = dataset_object.units['mpc']
    del dataset_object

    # Get shifted, tiled halo list.
    all_halo_x, all_halo_y, \
      all_halo_radius, all_halo_mass = \
      _make_slice_halo_list(slice, halo_list, virial_overdensity)

    # Construct list of halos
    halo_map = []

    for q in range(len(all_halo_x)):
        # Give radius in both physics units and
        # units of the image (0 to 1).
        radius_mpc = all_halo_radius[q] * Mpc_units
        radius_image = all_halo_radius[q] / slice['box_width_fraction']

        halo_map.append({'x': all_halo_x[q] / slice['box_width_fraction'],
                         'y': all_halo_y[q] / slice['box_width_fraction'],
                         'redshift': slice['redshift'],
                         'radius_mpc': radius_mpc,
                         'radius_image': radius_image,
                         'mass': all_halo_mass[q]})

    return halo_map

def _make_slice_halo_list(slice, halo_list, virial_overdensity):
    "Make shifted, tiled list of halos for halo mask and halo map."

   # Make numpy arrays for halo centers and virial radii.
    halo_x = []
    halo_y = []
    halo_depth = []
    halo_radius = []
    halo_mass = []

    # Get units to convert virial radii to code units.
    dataset_object = load(slice['filename'])
    Mpc_units = dataset_object.units['mpc']
    del dataset_object

    for halo in halo_list:
        if halo is not None:
            center = copy.deepcopy(halo['center'])
            halo_depth.append(center.pop(slice['projection_axis']))
            halo_x.append(center[0])
            halo_y.append(center[1])
            halo_radius.append(halo['RadiusMpc_%d' % virial_overdensity] /
                               Mpc_units)
            halo_mass.append(halo['TotalMassMsun_%d' % virial_overdensity])

    halo_x = np.array(halo_x)
    halo_y = np.array(halo_y)
    halo_depth = np.array(halo_depth)
    halo_radius = np.array(halo_radius)
    halo_mass = np.array(halo_mass)

    # Adjust halo centers along line of sight.
    depth_center = slice['projection_center'][slice['projection_axis']]
    depth_left = depth_center - 0.5 * slice['box_depth_fraction']
    depth_right = depth_center + 0.5 * slice['box_depth_fraction']

    # Make boolean mask to pick out centers in region along line of sight.
    # Halos near edges may wrap around to other side.
    add_left = (halo_depth + halo_radius) > 1 # should be box width
    add_right = (halo_depth - halo_radius) < 0

    halo_depth = np.concatenate([halo_depth,
                                 (halo_depth[add_left]-1),
                                 (halo_depth[add_right]+1)])
    halo_x = np.concatenate([halo_x, halo_x[add_left], halo_x[add_right]])
    halo_y = np.concatenate([halo_y, halo_y[add_left], halo_y[add_right]])
    halo_radius = np.concatenate([halo_radius,
                                  halo_radius[add_left],
                                  halo_radius[add_right]])
    halo_mass = np.concatenate([halo_mass,
                                halo_mass[add_left],
                                halo_mass[add_right]])

    del add_left, add_right

    # Cut out the halos outside the region of interest.
    if (slice['box_depth_fraction'] < 1):
        if (depth_left < 0):
            mask = ((halo_depth + halo_radius >= 0) &
                    (halo_depth - halo_radius <= depth_right)) | \
                ((halo_depth + halo_radius >= depth_left + 1) &
                 (halo_depth - halo_radius <= 1))
        elif (depth_right > 1):
            mask = ((halo_depth + halo_radius >= 0) &
                    (halo_depth - halo_radius <= depth_right - 1)) | \
                ((halo_depth + halo_radius >= depth_left) &
                 (halo_depth - halo_radius <= 1))
        else:
            mask = (halo_depth + halo_radius >= depth_left) & \
              (halo_depth - halo_radius <= depth_right)

        halo_x = halo_x[mask]
        halo_y = halo_y[mask]
        halo_radius = halo_radius[mask]
        halo_mass = halo_mass[mask]
        del mask
    del halo_depth

    all_halo_x = np.array([])
    all_halo_y = np.array([])
    all_halo_radius = np.array([])
    all_halo_mass = np.array([])

    # Tile halos of width box fraction is greater than one.
    # Copy original into offset positions to make tiles.
    for x in range(int(np.ceil(slice['box_width_fraction']))):
        for y in range(int(np.ceil(slice['box_width_fraction']))):
            all_halo_x = np.concatenate([all_halo_x, halo_x+x])
            all_halo_y = np.concatenate([all_halo_y, halo_y+y])
            all_halo_radius = np.concatenate([all_halo_radius, halo_radius])
            all_halo_mass = np.concatenate([all_halo_mass, halo_mass])

    del halo_x, halo_y, halo_radius, halo_mass

    # Shift centers laterally.
    offset = copy.deepcopy(slice['projection_center'])
    del offset[slice['projection_axis']]

    # Shift x and y positions.
    all_halo_x -= offset[0]
    all_halo_y -= offset[1]

    # Wrap off-edge centers back around to
    # other side (periodic boundary conditions).
    all_halo_x[all_halo_x < 0] += np.ceil(slice['box_width_fraction'])
    all_halo_y[all_halo_y < 0] += np.ceil(slice['box_width_fraction'])

    # After shifting, some centers have fractional coverage
    # on both sides of the box.
    # Find those centers and make copies to be placed on the other side.

    # Centers hanging off the right edge.
    add_x_right = all_halo_x + all_halo_radius > \
      np.ceil(slice['box_width_fraction'])
    add_x_halo_x = all_halo_x[add_x_right]
    add_x_halo_x -= np.ceil(slice['box_width_fraction'])
    add_x_halo_y = all_halo_y[add_x_right]
    add_x_halo_radius = all_halo_radius[add_x_right]
    add_x_halo_mass = all_halo_mass[add_x_right]
    del add_x_right

    # Centers hanging off the left edge.
    add_x_left = all_halo_x - all_halo_radius < 0
    add2_x_halo_x = all_halo_x[add_x_left]
    add2_x_halo_x += np.ceil(slice['box_width_fraction'])
    add2_x_halo_y = all_halo_y[add_x_left]
    add2_x_halo_radius = all_halo_radius[add_x_left]
    add2_x_halo_mass = all_halo_mass[add_x_left]
    del add_x_left

    # Centers hanging off the top edge.
    add_y_right = all_halo_y + all_halo_radius > \
      np.ceil(slice['box_width_fraction'])
    add_y_halo_x = all_halo_x[add_y_right]
    add_y_halo_y = all_halo_y[add_y_right]
    add_y_halo_y -= np.ceil(slice['box_width_fraction'])
    add_y_halo_radius = all_halo_radius[add_y_right]
    add_y_halo_mass = all_halo_mass[add_y_right]
    del add_y_right

    # Centers hanging off the bottom edge.
    add_y_left = all_halo_y - all_halo_radius < 0
    add2_y_halo_x = all_halo_x[add_y_left]
    add2_y_halo_y = all_halo_y[add_y_left]
    add2_y_halo_y += np.ceil(slice['box_width_fraction'])
    add2_y_halo_radius = all_halo_radius[add_y_left]
    add2_y_halo_mass = all_halo_mass[add_y_left]
    del add_y_left

    # Add the hanging centers back to the projection data.
    all_halo_x = np.concatenate([all_halo_x,
                                 add_x_halo_x, add2_x_halo_x,
                                 add_y_halo_x, add2_y_halo_x])
    all_halo_y = np.concatenate([all_halo_y,
                                 add_x_halo_y, add2_x_halo_y,
                                 add_y_halo_y, add2_y_halo_y])
    all_halo_radius = np.concatenate([all_halo_radius,
                                      add_x_halo_radius,
                                      add2_x_halo_radius,
                                      add_y_halo_radius,
                                      add2_y_halo_radius])
    all_halo_mass = np.concatenate([all_halo_mass,
                                    add_x_halo_mass,
                                    add2_x_halo_mass,
                                    add_y_halo_mass,
                                    add2_y_halo_mass])

    del add_x_halo_x, add_x_halo_y, add_x_halo_radius
    del add2_x_halo_x, add2_x_halo_y, add2_x_halo_radius
    del add_y_halo_x, add_y_halo_y, add_y_halo_radius
    del add2_y_halo_x, add2_y_halo_y, add2_y_halo_radius

    # Cut edges to proper width.
    cut_mask = (all_halo_x - all_halo_radius <
                slice['box_width_fraction']) & \
        (all_halo_y - all_halo_radius <
         slice['box_width_fraction'])
    all_halo_x = all_halo_x[cut_mask]
    all_halo_y = all_halo_y[cut_mask]
    all_halo_radius = all_halo_radius[cut_mask]
    all_halo_mass = all_halo_mass[cut_mask]
    del cut_mask

    return (all_halo_x, all_halo_y,
            all_halo_radius, all_halo_mass)



"""
Orion-specific fields

Author: J. S. Oishi <jsoishi@gmail.com>
Affiliation: UC Berkeley
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 J. S. Oishi, Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from yt.utilities.physical_constants import \
    mh, kboltz
from yt.data_objects.field_info_container import \
    FieldInfoContainer, \
    NullFunc, \
    TranslationFunc, \
    FieldInfo, \
    ValidateParameter, \
    ValidateDataField, \
    ValidateProperty, \
    ValidateSpatial, \
    ValidateGridType
import yt.data_objects.universal_fields


KnownOrionFields = FieldInfoContainer()
add_orion_field = KnownOrionFields.add_field

OrionFieldInfo = FieldInfoContainer.create_with_fallback(FieldInfo)
add_field = OrionFieldInfo.add_field

add_orion_field("density", function=NullFunc, take_log=True,
                validators = [ValidateDataField("density")],
                units=r"\rm{g}/\rm{cm}^3")
KnownOrionFields["density"]._projected_units =r"\rm{g}/\rm{cm}^2"

add_orion_field("eden", function=NullFunc, take_log=True,
                validators = [ValidateDataField("eden")],
                units=r"\rm{erg}/\rm{cm}^3")

add_orion_field("xmom", function=NullFunc, take_log=False,
                validators = [ValidateDataField("xmom")],
                units=r"\rm{g}/\rm{cm^2\ s}")

add_orion_field("ymom", function=NullFunc, take_log=False,
                validators = [ValidateDataField("ymom")],
                units=r"\rm{gm}/\rm{cm^2\ s}")

add_orion_field("zmom", function=NullFunc, take_log=False,
                validators = [ValidateDataField("zmom")],
                units=r"\rm{g}/\rm{cm^2\ s}")

translation_dict = {"x-velocity": "xvel",
                    "y-velocity": "yvel",
                    "z-velocity": "zvel",
                    "Density": "density",
                    "TotalEnergy": "eden",
                    "Temperature": "temperature",
                    "x-momentum": "xmom",
                    "y-momentum": "ymom",
                    "z-momentum": "zmom"
                   }

for f,v in translation_dict.items():
    if v not in KnownOrionFields:
        add_orion_field(v, function=NullFunc, take_log=False,
                  validators = [ValidateDataField(v)])
    ff = KnownOrionFields[v]
    add_field(f, TranslationFunc(v),
              take_log=KnownOrionFields[v].take_log,
              units = ff._units, display_name=f)

def _xVelocity(field, data):
    """generate x-velocity from x-momentum and density
    
    """
    return data["xmom"]/data["density"]
add_orion_field("x-velocity",function=_xVelocity, take_log=False,
                units=r'\rm{cm}/\rm{s}')

def _yVelocity(field,data):
    """generate y-velocity from y-momentum and density

    """
    #try:
    #    return data["xvel"]
    #except KeyError:
    return data["ymom"]/data["density"]
add_orion_field("y-velocity",function=_yVelocity, take_log=False,
                units=r'\rm{cm}/\rm{s}')

def _zVelocity(field,data):
    """generate z-velocity from z-momentum and density
    
    """
    return data["zmom"]/data["density"]
add_orion_field("z-velocity",function=_zVelocity, take_log=False,
                units=r'\rm{cm}/\rm{s}')

def _ThermalEnergy(field, data):
    """generate thermal (gas energy). Dual Energy Formalism was
        implemented by Stella, but this isn't how it's called, so I'll
        leave that commented out for now.
    """
    #if data.pf["DualEnergyFormalism"]:
    #    return data["GasEnergy"]
    #else:
    return data["TotalEnergy"] - 0.5 * data["density"] * (
        data["x-velocity"]**2.0
        + data["y-velocity"]**2.0
        + data["z-velocity"]**2.0 )
add_field("ThermalEnergy", function=_ThermalEnergy,
                units=r"\rm{ergs}/\rm{cm^3}")

def _Pressure(field,data):
    """M{(Gamma-1.0)*e, where e is thermal energy density
       NB: this will need to be modified for radiation
    """
    return (data.pf["Gamma"] - 1.0)*data["ThermalEnergy"]
add_field("Pressure", function=_Pressure, units=r"\rm{dyne}/\rm{cm}^{2}")

def _Temperature(field,data):
    return (data.pf["Gamma"]-1.0)*data.pf["mu"]*mh*data["ThermalEnergy"]/(kboltz*data["Density"])
add_field("Temperature",function=_Temperature,units=r"\rm{Kelvin}",take_log=False)

# particle fields

def particle_func(p_field, dtype='float64'):
    def _Particles(field, data):
        io = data.hierarchy.io
        if not data.NumberOfParticles > 0:
            return np.array([], dtype=dtype)
        else:
            return io._read_particles(data, p_field).astype(dtype)

    return _Particles

_particle_field_list = ["mass", 
                        "position_x",
                        "position_y",
                        "position_z",
                        "momentum_x",
                        "momentum_y",
                        "momentum_z",
                        "angmomen_x",
                        "angmomen_y",
                        "angmomen_z",
                        "mlast",
                        "r",
                        "mdeut",
                        "n",
                        "mdot",
                        "burnstate",
                        "luminosity",
                        "id"]

for pf in _particle_field_list:
    pfunc = particle_func("particle_%s" % (pf))
    add_field("particle_%s" % pf, function=pfunc,
              validators = [ValidateSpatial(0)],
              particle_type=True)

def _ParticleMass(field, data):
    particles = data["particle_mass"].astype('float64')
    return particles

def _ParticleMassMsun(field, data):
    particles = data["particle_mass"].astype('float64')
    return particles/1.989e33

add_field("ParticleMass",
          function=_ParticleMass, validators=[ValidateSpatial(0)],
          particle_type=True)
add_field("ParticleMassMsun",
          function=_ParticleMassMsun, validators=[ValidateSpatial(0)],
          particle_type=True)



"""
Data structures for Chombo.

Author: Matthew Turk <matthewturk@gmail.com>
Author: J. S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 Matthew Turk, J. S. Oishi.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import h5py
import re
import os
import weakref
import numpy as np

from collections import \
     defaultdict
from string import \
     strip, \
     rstrip
from stat import \
     ST_CTIME

from .definitions import \
     chombo2enzoDict, \
     yt2chomboFieldsDict, \
     parameterDict \

from yt.funcs import *
from yt.data_objects.grid_patch import \
     AMRGridPatch
from yt.data_objects.hierarchy import \
     AMRHierarchy
from yt.data_objects.static_output import \
     StaticOutput
from yt.utilities.definitions import \
     mpc_conversion, sec_conversion
from yt.utilities.parallel_tools.parallel_analysis_interface import \
     parallel_root_only
from yt.utilities.io_handler import \
    io_registry

from yt.data_objects.field_info_container import \
    FieldInfoContainer, NullFunc
from .fields import ChomboFieldInfo, KnownChomboFields

class ChomboGrid(AMRGridPatch):
    _id_offset = 0
    __slots__ = ["_level_id", "stop_index"]
    def __init__(self, id, hierarchy, level, start, stop):
        AMRGridPatch.__init__(self, id, filename = hierarchy.hierarchy_filename,
                              hierarchy = hierarchy)
        self.Parent = []
        self.Children = []
        self.Level = level
        self.ActiveDimensions = stop - start + 1

    def get_global_startindex(self):
        """
        Return the integer starting index for each dimension at the current
        level.

        """
        if self.start_index != None:
            return self.start_index
        if self.Parent == []:
            iLE = self.LeftEdge - self.pf.domain_left_edge
            start_index = iLE / self.dds
            return np.rint(start_index).astype('int64').ravel()
        pdx = self.Parent[0].dds
        start_index = (self.Parent[0].get_global_startindex()) + \
            np.rint((self.LeftEdge - self.Parent[0].LeftEdge)/pdx)
        self.start_index = (start_index*self.pf.refine_by).astype('int64').ravel()
        return self.start_index

    def _setup_dx(self):
        # has already been read in and stored in hierarchy
        self.dds = self.hierarchy.dds_list[self.Level]
        self.field_data['dx'], self.field_data['dy'], self.field_data['dz'] = self.dds

class ChomboHierarchy(AMRHierarchy):

    grid = ChomboGrid

    def __init__(self,pf,data_style='chombo_hdf5'):
        self.domain_left_edge = pf.domain_left_edge
        self.domain_right_edge = pf.domain_right_edge
        self.data_style = data_style
        self.field_indexes = {}
        self.parameter_file = weakref.proxy(pf)
        # for now, the hierarchy file is the parameter file!
        self.hierarchy_filename = os.path.abspath(
            self.parameter_file.parameter_filename)
        self.directory = pf.fullpath
        self._handle = pf._handle

        self.float_type = self._handle['/level_0']['data:datatype=0'].dtype.name
        self._levels = self._handle.keys()[1:]
        AMRHierarchy.__init__(self,pf,data_style)
        self._read_particles()

    def _read_particles(self):
        self.particle_filename = self.hierarchy_filename[:-4] + 'sink'
        if not os.path.exists(self.particle_filename): return
        with open(self.particle_filename, 'r') as f:
            lines = f.readlines()
            self.num_stars = int(lines[0].strip().split(' ')[0])
            for line in lines[1:]:
                particle_position_x = float(line.split(' ')[1])
                particle_position_y = float(line.split(' ')[2])
                particle_position_z = float(line.split(' ')[3])
                coord = [particle_position_x, particle_position_y, particle_position_z]
                # for each particle, determine which grids contain it
                # copied from object_finding_mixin.py
                mask=np.ones(self.num_grids)
                for i in xrange(len(coord)):
                    np.choose(np.greater(self.grid_left_edge[:,i],coord[i]), (mask,0), mask)
                    np.choose(np.greater(self.grid_right_edge[:,i],coord[i]), (0,mask), mask)
                ind = np.where(mask == 1)
                selected_grids = self.grids[ind]
                # in orion, particles always live on the finest level.
                # so, we want to assign the particle to the finest of
                # the grids we just found
                if len(selected_grids) != 0:
                    grid = sorted(selected_grids, key=lambda grid: grid.Level)[-1]
                    ind = np.where(self.grids == grid)[0][0]
                    self.grid_particle_count[ind] += 1
                    self.grids[ind].NumberOfParticles += 1

    def _detect_fields(self):
        ncomp = int(self._handle['/'].attrs['num_components'])
        self.field_list = [c[1] for c in self._handle['/'].attrs.items()[-ncomp:]]
          
    def _setup_classes(self):
        dd = self._get_data_reader_dict()
        AMRHierarchy._setup_classes(self, dd)
        self.object_types.sort()

    def _count_grids(self):
        self.num_grids = 0
        for lev in self._levels:
            self.num_grids += self._handle[lev]['Processors'].len()

    def _parse_hierarchy(self):
        f = self._handle # shortcut

        # this relies on the first Group in the H5 file being
        # 'Chombo_global'
        levels = f.keys()[1:]
        grids = []
        self.dds_list = []
        i = 0
        for lev in levels:
            level_number = int(re.match('level_(\d+)',lev).groups()[0])
            boxes = f[lev]['boxes'].value
            dx = f[lev].attrs['dx']
            self.dds_list.append(dx * np.ones(3))
            for level_id, box in enumerate(boxes):
                si = np.array([box['lo_%s' % ax] for ax in 'ijk'])
                ei = np.array([box['hi_%s' % ax] for ax in 'ijk'])
                pg = self.grid(len(grids),self,level=level_number,
                               start = si, stop = ei)
                grids.append(pg)
                grids[-1]._level_id = level_id
                self.grid_left_edge[i] = dx*si.astype(self.float_type)
                self.grid_right_edge[i] = dx*(ei.astype(self.float_type)+1)
                self.grid_particle_count[i] = 0
                self.grid_dimensions[i] = ei - si + 1
                i += 1
        self.grids = np.empty(len(grids), dtype='object')
        for gi, g in enumerate(grids): self.grids[gi] = g
#        self.grids = np.array(self.grids, dtype='object')

    def _populate_grid_objects(self):
        for g in self.grids:
            g._prepare_grid()
            g._setup_dx()

        for g in self.grids:
            g.Children = self._get_grid_children(g)
            for g1 in g.Children:
                g1.Parent.append(g)
        self.max_level = self.grid_levels.max()

    def _setup_derived_fields(self):
        self.derived_field_list = []

    def _get_grid_children(self, grid):
        mask = np.zeros(self.num_grids, dtype='bool')
        grids, grid_ind = self.get_box_grids(grid.LeftEdge, grid.RightEdge)
        mask[grid_ind] = True
        return [g for g in self.grids[mask] if g.Level == grid.Level + 1]

    def _setup_data_io(self):
        self.io = io_registry[self.data_style](self.parameter_file)

class ChomboStaticOutput(StaticOutput):
    _hierarchy_class = ChomboHierarchy
    _fieldinfo_fallback = ChomboFieldInfo
    _fieldinfo_known = KnownChomboFields

    def __init__(self, filename, data_style='chombo_hdf5',
                 storage_filename = None, ini_filename = None):
        self._handle = h5py.File(filename,'r')
        self.current_time = self._handle.attrs['time']
        self.ini_filename = ini_filename
        self.fullplotdir = os.path.abspath(filename)
        StaticOutput.__init__(self,filename,data_style)
        self.storage_filename = storage_filename
        self.cosmological_simulation = False

        # These are parameters that I very much wish to get rid of.
        self.parameters["HydroMethod"] = 'chombo' # always PPM DE
        self.parameters["DualEnergyFormalism"] = 0 
        self.parameters["EOSType"] = -1 # default

    def __del__(self):
        self._handle.close()

    def _set_units(self):
        """
        Generates the conversion to various physical _units based on the parameter file
        """
        self.units = {}
        self.time_units = {}
        if len(self.parameters) == 0:
            self._parse_parameter_file()
        self._setup_nounits_units()
        self.conversion_factors = defaultdict(lambda: 1.0)
        self.time_units['1'] = 1
        self.units['1'] = 1.0
        self.units['unitary'] = 1.0 / (self.domain_right_edge - self.domain_left_edge).max()
        seconds = 1 #self["Time"]
        for unit in sec_conversion.keys():
            self.time_units[unit] = seconds / sec_conversion[unit]
        for key in yt2chomboFieldsDict:
            self.conversion_factors[key] = 1.0

    def _setup_nounits_units(self):
        z = 0
        mylog.warning("Setting 1.0 in code units to be 1.0 cm")
        if not self.has_key("TimeUnits"):
            mylog.warning("No time units.  Setting 1.0 = 1 second.")
            self.conversion_factors["Time"] = 1.0
        for unit in mpc_conversion.keys():
            self.units[unit] = mpc_conversion[unit] / mpc_conversion["cm"]


    def _localize(self, f, default):
        if f is None:
            return os.path.join(self.directory, default)
        return f

    def _parse_parameter_file(self):
        """
        Check to see whether an 'orion2.ini' file
        exists in the plot file directory. If one does, attempt to parse it.
        Otherwise grab the dimensions from the hdf5 file.
        """
        
        if os.path.isfile('orion2.ini'): self._parse_inputs_file('orion2.ini')
        self.unique_identifier = \
                               int(os.stat(self.parameter_filename)[ST_CTIME])
        self.domain_left_edge = self.__calc_left_edge()
        self.domain_right_edge = self.__calc_right_edge()
        self.domain_dimensions = self.__calc_domain_dimensions()
        self.dimensionality = 3
        self.refine_by = self._handle['/level_0'].attrs['ref_ratio']
        self.periodicity = (True, True, True)

    def _parse_inputs_file(self, ini_filename):
        self.fullplotdir = os.path.abspath(self.parameter_filename)
        self.ini_filename = self._localize( \
            self.ini_filename, ini_filename)
        self.unique_identifier = \
                               int(os.stat(self.parameter_filename)[ST_CTIME])
        lines = open(self.ini_filename).readlines()
        # read the file line by line, storing important parameters
        for lineI, line in enumerate(lines):
            try:
                param, sep, vals = map(rstrip,line.partition(' '))
            except ValueError:
                mylog.error("ValueError: '%s'", line)
            if chombo2enzoDict.has_key(param):
                paramName = chombo2enzoDict[param]
                t = map(parameterDict[paramName], vals.split())
                if len(t) == 1:
                    self.parameters[paramName] = t[0]
                else:
                    if paramName == "RefineBy":
                        self.parameters[paramName] = t[0]
                    else:
                        self.parameters[paramName] = t

    def __calc_left_edge(self):
        fileh = self._handle
        dx0 = fileh['/level_0'].attrs['dx']
        LE = dx0*((np.array(list(fileh['/level_0'].attrs['prob_domain'])))[0:3])
        return LE

    def __calc_right_edge(self):
        fileh = self._handle
        dx0 = fileh['/level_0'].attrs['dx']
        RE = dx0*((np.array(list(fileh['/level_0'].attrs['prob_domain'])))[3:] + 1)
        return RE

    def __calc_domain_dimensions(self):
        fileh = self._handle
        L_index = ((np.array(list(fileh['/level_0'].attrs['prob_domain'])))[0:3])
        R_index = ((np.array(list(fileh['/level_0'].attrs['prob_domain'])))[3:] + 1)
        return R_index - L_index

    @classmethod
    def _is_valid(self, *args, **kwargs):
        if not os.path.isfile('pluto.ini'):
            try:
                fileh = h5py.File(args[0],'r')
                valid = "Chombo_global" in fileh["/"]
                fileh.close()
                return valid
            except:
                pass
        return False

    @parallel_root_only
    def print_key_parameters(self):
        for a in ["current_time", "domain_dimensions", "domain_left_edge",
                  "domain_right_edge"]:
            if not hasattr(self, a):
                mylog.error("Missing %s in parameter file definition!", a)
                continue
            v = getattr(self, a)
            mylog.info("Parameters: %-25s = %s", a, v)



"""
Maestro data-file handling functions - identical to Orion frontend,
but with all instances of "orion" changed to "maestro".

Author: Matthew Turk <matthewturk@gmail.com>
Author: J. S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Author: Chris Malone <chris.m.malone@gmail.com>
Affiliation: SUNY Stony Brook
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import numpy as np
from yt.utilities.io_handler import \
           BaseIOHandler

from definitions import \
    yt2maestroFieldsDict

class IOHandlerNative(BaseIOHandler):

    _data_style = "maestro"

    def modify(self, field):
        return field.swapaxes(0,2)

    def _read_data(self,grid,field):
        """
        reads packed multiFABs output by BoxLib in "NATIVE" format.

        """
        filen = os.path.expanduser(grid.filename[field])
        off = grid._offset[field]
        inFile = open(filen,'rb')
        inFile.seek(off)
        header = inFile.readline()
        header.strip()

        if grid._paranoid:
            mylog.warn("Maestro Native reader: Paranoid read mode.")
            headerRe = re.compile(maestro_FAB_header_pattern)
            bytesPerReal,endian,start,stop,centerType,nComponents = headerRe.search(header).groups()

            # we will build up a dtype string, starting with endian
            # check endianness (this code is ugly. fix?)
            bytesPerReal = int(bytesPerReal)
            if bytesPerReal == int(endian[0]):
                dtype = '<'
            elif bytesPerReal == int(endian[-1]):
                dtype = '>'
            else:
                raise ValueError("FAB header is neither big nor little endian. Perhaps the file is corrupt?")

            dtype += ('f%i'% bytesPerReal) #always a floating point

            # determine size of FAB
            start = np.array(map(int,start.split(',')))
            stop = np.array(map(int,stop.split(',')))

            gridSize = stop - start + 1

            error_count = 0
            if (start != grid.start).any():
                print "Paranoia Error: Cell_H and %s do not agree on grid start." %grid.filename
                error_count += 1
            if (stop != grid.stop).any():
                print "Paranoia Error: Cell_H and %s do not agree on grid stop." %grid.filename
                error_count += 1
            if (gridSize != grid.ActiveDimensions).any():
                print "Paranoia Error: Cell_H and %s do not agree on grid dimensions." %grid.filename
                error_count += 1
            if bytesPerReal != grid.hierarchy._bytesPerReal:
                print "Paranoia Error: Cell_H and %s do not agree on bytes per real number." %grid.filename
                error_count += 1
            if (bytesPerReal == grid.hierarchy._bytesPerReal and dtype != grid.hierarchy._dtype):
                print "Paranoia Error: Cell_H and %s do not agree on endianness." %grid.filename
                error_count += 1

            if error_count > 0:
                raise RunTimeError("Paranoia unveiled %i differences between Cell_H and %s." % (error_count, grid.filename))

        else:
            start = grid.start_index
            stop = grid.stop_index
            dtype = grid.hierarchy._dtype
            bytesPerReal = grid.hierarchy._bytesPerReal

        nElements = grid.ActiveDimensions.prod()

        # one field has nElements*bytesPerReal bytes and is located
        # nElements*bytesPerReal*field_index from the offset location
        if yt2maestroFieldsDict.has_key(field):
            fieldname = yt2maestroFieldsDict[field]
        else:
            fieldname = field
        field_index = grid.field_indexes[fieldname]
        inFile.seek(int(nElements*bytesPerReal*field_index),1)
        field = np.fromfile(inFile,count=nElements,dtype=dtype)
        field = field.reshape(grid.ActiveDimensions[::-1]).swapaxes(0,2)

        # we can/should also check against the max and min in the header file

        inFile.close()
        return field




"""
Maestro-specific fields - borrows heavily from Orion frontend.

Author: J. S. Oishi <jsoishi@gmail.com>
Affiliation: UC Berkeley
Author: Chris Malone <chris.m.malone@gmail.com>
Affiliation: SUNY Stony Brook
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 J. S. Oishi, Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from yt.utilities.physical_constants import \
    mh, kboltz
from yt.data_objects.field_info_container import \
    FieldInfoContainer, \
    FieldInfo, \
    ValidateParameter, \
    ValidateDataField, \
    ValidateProperty, \
    ValidateSpatial, \
    ValidateGridType
import yt.data_objects.universal_fields

KnownMaestroFields = FieldInfoContainer()
add_maestro_field = KnownMaestroFields.add_field

MaestroFieldInfo = FieldInfoContainer.create_with_fallback(FieldInfo)
add_field = MaestroFieldInfo.add_field

add_field("density", function=lambda a,b: None, take_log=True,
          validators = [ValidateDataField("density")],
          units=r"\rm{g}/\rm{cm}^3")
MaestroFieldInfo["density"]._projected_units =r"\rm{g}/\rm{cm}^2"

translation_dict = {"x-velocity": "x_vel",
                    "y-velocity": "y_vel",
                    "z-velocity": "z_vel",
                    "Density": "density",
                    "Temperature": "tfromp"
                   }

def _generate_translation(mine, theirs):
    add_field(theirs, function=lambda a, b: b[mine], take_log=True)

for f,v in translation_dict.items():
    if v not in MaestroFieldInfo:
        add_field(v, function=lambda a,b: None, take_log=False,
                  validators = [ValidateDataField(v)])
#    print "Setting up translator from %s to %s" % (v, f)
    _generate_translation(v, f)




"""
RAMSES-specific data structures

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import stat
import weakref

from yt.funcs import *
from yt.data_objects.grid_patch import \
      AMRGridPatch
from yt.data_objects.hierarchy import \
      AMRHierarchy
from yt.data_objects.static_output import \
      StaticOutput

try:
    import _ramses_reader
except ImportError:
    _ramses_reader = None
from .fields import RAMSESFieldInfo, KnownRAMSESFields
from yt.utilities.definitions import \
    mpc_conversion, sec_conversion
from yt.utilities.lib import \
    get_box_grids_level
from yt.utilities.io_handler import \
    io_registry
from yt.data_objects.field_info_container import \
    FieldInfoContainer, NullFunc

def num_deep_inc(f):
    def wrap(self, *args, **kwargs):
        self.num_deep += 1
        rv = f(self, *args, **kwargs)
        self.num_deep -= 1
        return rv
    return wrap

class RAMSESGrid(AMRGridPatch):
    _id_offset = 0
    #__slots__ = ["_level_id", "stop_index"]
    def __init__(self, id, hierarchy, level, locations, start_index):
        AMRGridPatch.__init__(self, id, filename = hierarchy.hierarchy_filename,
                              hierarchy = hierarchy)
        self.Level = level
        self.Parent = []
        self.Children = []
        self.locations = locations
        self.domain = locations[0,0]
        self.start_index = start_index.copy()

    def _setup_dx(self):
        # So first we figure out what the index is.  We don't assume
        # that dx=dy=dz , at least here.  We probably do elsewhere.
        id = self.id - self._id_offset
        if len(self.Parent) > 0:
            self.dds = self.Parent[0].dds / self.pf.refine_by
        else:
            LE, RE = self.hierarchy.grid_left_edge[id,:], \
                     self.hierarchy.grid_right_edge[id,:]
            self.dds = np.array((RE-LE)/self.ActiveDimensions)
        if self.pf.dimensionality < 2: self.dds[1] = 1.0
        if self.pf.dimensionality < 3: self.dds[2] = 1.0
        self.field_data['dx'], self.field_data['dy'], self.field_data['dz'] = self.dds

    def get_global_startindex(self):
        """
        Return the integer starting index for each dimension at the current
        level.
        """
        if self.start_index != None:
            return self.start_index
        if len(self.Parent) == 0:
            start_index = self.LeftEdge / self.dds
            return np.rint(start_index).astype('int64').ravel()
        pdx = self.Parent[0].dds
        start_index = (self.Parent[0].get_global_startindex()) + \
                       np.rint((self.LeftEdge - self.Parent[0].LeftEdge)/pdx)
        self.start_index = (start_index*self.pf.refine_by).astype('int64').ravel()
        return self.start_index

    def __repr__(self):
        return "RAMSESGrid_%04i (%s)" % (self.id, self.ActiveDimensions)

class RAMSESHierarchy(AMRHierarchy):

    grid = RAMSESGrid
    _handle = None
    
    def __init__(self,pf,data_style='ramses'):
        self.data_style = data_style
        self.parameter_file = weakref.proxy(pf)
        # for now, the hierarchy file is the parameter file!
        self.hierarchy_filename = self.parameter_file.parameter_filename
        self.directory = os.path.dirname(self.hierarchy_filename)
        self.tree_proxy = pf.ramses_tree

        self.float_type = np.float64
        AMRHierarchy.__init__(self,pf,data_style)

    def _initialize_data_storage(self):
        pass

    def _detect_fields(self):
        self.field_list = self.tree_proxy.field_names[:]
    
    def _setup_field_list(self):
        if self.parameter_file.use_particles:
            # We know which particle fields will exist -- pending further
            # changes in the future.
            for field in art_particle_field_names:
                def external_wrapper(f):
                    def _convert_function(data):
                        return data.convert(f)
                    return _convert_function
                cf = external_wrapper(field)
                # Note that we call add_field on the field_info directly.  This
                # will allow the same field detection mechanism to work for 1D,
                # 2D and 3D fields.
                self.pf.field_info.add_field(field, NullFunc,
                                             convert_function=cf,
                                             take_log=False, particle_type=True)

    def _setup_classes(self):
        dd = self._get_data_reader_dict()
        AMRHierarchy._setup_classes(self, dd)
        self.object_types.sort()

    def _count_grids(self):
        # We have to do all the patch-coalescing here.
        LEVEL_OF_EDGE = 7
        MAX_EDGE = (2 << (LEVEL_OF_EDGE- 1))
        level_info = self.tree_proxy.count_zones()
        num_ogrids = sum(level_info)
        ogrid_left_edge = np.zeros((num_ogrids,3), dtype='float64')
        ogrid_right_edge = np.zeros((num_ogrids,3), dtype='float64')
        ogrid_levels = np.zeros((num_ogrids,1), dtype='int32')
        ogrid_file_locations = np.zeros((num_ogrids,6), dtype='int64')
        ogrid_hilbert_indices = np.zeros(num_ogrids, dtype='uint64')
        ochild_masks = np.zeros((num_ogrids, 8), dtype='int32')
        self.tree_proxy.fill_hierarchy_arrays(
            self.pf.domain_dimensions,
            ogrid_left_edge, ogrid_right_edge,
            ogrid_levels, ogrid_file_locations, 
            ogrid_hilbert_indices, ochild_masks,
            self.pf.domain_left_edge, self.pf.domain_right_edge)
        #insert_ipython()
        # Now we can rescale
        mi, ma = ogrid_left_edge.min(), ogrid_right_edge.max()
        DL = self.pf.domain_left_edge
        DR = self.pf.domain_right_edge
        DW = DR - DL
        ogrid_left_edge = (ogrid_left_edge - mi)/(ma - mi) * (DR - DL) + DL
        ogrid_right_edge = (ogrid_right_edge - mi)/(ma - mi) * (DR - DL) + DL
        #import pdb;pdb.set_trace()
        # We now have enough information to run the patch coalescing 
        self.proto_grids = []
        for level in xrange(len(level_info)):
            if level_info[level] == 0: continue
            # Get the indices of grids on this level
            ggi = (ogrid_levels == level).ravel()
            dims = np.ones((ggi.sum(), 3), dtype='int64') * 2 
            mylog.info("Re-gridding level %s: %s octree grids", level, ggi.sum())
            nd = self.pf.domain_dimensions * 2**level
            fl = ogrid_file_locations[ggi,:]
            # Now our initial protosubgrid
            #if level == 6: raise RuntimeError
            # We want grids that cover no more than MAX_EDGE cells in every direction
            psgs = []
            # left_index is integers of the index, with respect to this level
            left_index = np.rint((ogrid_left_edge[ggi,:]) * nd / DW ).astype('int64')
            # we've got octs, so it's +2
            pbar = get_pbar("Re-gridding ", left_index.shape[0])
            dlp = [None, None, None]
            i = 0
            # We now calculate the hilbert curve position of every left_index,
            # of the octs, with respect to a lower order hilbert curve.
            left_index_gridpatch = left_index >> LEVEL_OF_EDGE
            order = max(level + 1 - LEVEL_OF_EDGE, 0)
            # I'm not sure the best way to do this.
            hilbert_indices = _ramses_reader.get_hilbert_indices(order, left_index_gridpatch)
            #print level, hilbert_indices.min(), hilbert_indices.max()
            # Strictly speaking, we don't care about the index of any
            # individual oct at this point.  So we can then split them up.
            unique_indices = np.unique(hilbert_indices)
            mylog.debug("Level % 2i has % 10i unique indices for %0.3e octs",
                        level, unique_indices.size, hilbert_indices.size)
            locs, lefts = _ramses_reader.get_array_indices_lists(
                        hilbert_indices, unique_indices, left_index, fl)
            for ddleft_index, ddfl in zip(lefts, locs):
                for idomain in np.unique(ddfl[:,0]):
                    dom_ind = ddfl[:,0] == idomain
                    dleft_index = ddleft_index[dom_ind,:]
                    dfl = ddfl[dom_ind,:]
                    initial_left = np.min(dleft_index, axis=0)
                    idims = (np.max(dleft_index, axis=0) - initial_left).ravel()+2
                    psg = _ramses_reader.ProtoSubgrid(initial_left, idims,
                                    dleft_index, dfl)
                    if psg.efficiency <= 0: continue
                    self.num_deep = 0
                    psgs.extend(_ramses_reader.recursive_patch_splitting(
                        psg, idims, initial_left, 
                        dleft_index, dfl))
            mylog.debug("Done with level % 2i", level)
            pbar.finish()
            self.proto_grids.append(psgs)
            print sum(len(psg.grid_file_locations) for psg in psgs)
            sums = np.zeros(3, dtype='int64')
            mylog.info("Final grid count: %s", len(self.proto_grids[level]))
            if len(self.proto_grids[level]) == 1: continue
            #for g in self.proto_grids[level]:
            #    sums += [s.sum() for s in g.sigs]
            #assert(np.all(sums == dims.prod(axis=1).sum()))
        self.num_grids = sum(len(l) for l in self.proto_grids)

    def _parse_hierarchy(self):
        # We have important work to do
        grids = []
        gi = 0
        DL, DR = self.pf.domain_left_edge, self.pf.domain_right_edge
        DW = DR - DL
        for level, grid_list in enumerate(self.proto_grids):
            for g in grid_list:
                fl = g.grid_file_locations
                props = g.get_properties()
                self.grid_left_edge[gi,:] = (props[0,:] / (2.0**(level+1))) * DW + DL
                self.grid_right_edge[gi,:] = (props[1,:] / (2.0**(level+1))) * DW + DL
                self.grid_dimensions[gi,:] = props[2,:]
                self.grid_levels[gi,:] = level
                grids.append(self.grid(gi, self, level, fl, props[0,:]))
                gi += 1
        self.proto_grids = []
        self.grids = np.empty(len(grids), dtype='object')
        for gi, g in enumerate(grids): self.grids[gi] = g

    def _populate_grid_objects(self):
        mask = np.empty(self.grids.size, dtype='int32')
        print self.grid_levels.dtype
        for gi,g in enumerate(self.grids):
            get_box_grids_level(self.grid_left_edge[gi,:],
                                self.grid_right_edge[gi,:],
                                g.Level - 1,
                                self.grid_left_edge, self.grid_right_edge,
                                self.grid_levels, mask)
            parents = self.grids[mask.astype("bool")]
            if len(parents) > 0:
                g.Parent.extend((p for p in parents.tolist()
                        if p.locations[0,0] == g.locations[0,0]))
                for p in parents: p.Children.append(g)
            # Now we do overlapping siblings; note that one has to "win" with
            # siblings, so we assume the lower ID one will "win"
            get_box_grids_level(self.grid_left_edge[gi,:],
                                self.grid_right_edge[gi,:],
                                g.Level,
                                self.grid_left_edge, self.grid_right_edge,
                                self.grid_levels, mask, gi)
            mask[gi] = False
            siblings = self.grids[mask.astype("bool")]
            if len(siblings) > 0:
                g.OverlappingSiblings = siblings.tolist()
            g._prepare_grid()
            g._setup_dx()
        self.max_level = self.grid_levels.max()

    def _setup_derived_fields(self):
        self.derived_field_list = []

    def _setup_data_io(self):
        self.io = io_registry[self.data_style](self.tree_proxy)

class RAMSESStaticOutput(StaticOutput):
    _hierarchy_class = RAMSESHierarchy
    _fieldinfo_fallback = RAMSESFieldInfo
    _fieldinfo_known = KnownRAMSESFields
    _handle = None
    
    def __init__(self, filename, data_style='ramses',
                 storage_filename = None):
        # Here we want to initiate a traceback, if the reader is not built.
        import _ramses_reader
        StaticOutput.__init__(self, filename, data_style)
        self.storage_filename = storage_filename

    def __repr__(self):
        return self.basename.rsplit(".", 1)[0]
        
    def _set_units(self):
        """
        Generates the conversion to various physical _units based on the parameter file
        """
        self.units = {}
        self.time_units = {}
        if len(self.parameters) == 0:
            self._parse_parameter_file()
        self._setup_nounits_units()
        self.conversion_factors = defaultdict(lambda: 1.0)
        self.time_units['1'] = 1
        self.units['1'] = 1.0
        self.units['unitary'] = 1.0 / (self.domain_right_edge - self.domain_left_edge).max()
        self.conversion_factors["Density"] = self.parameters['unit_d']
        vel_u = self.parameters['unit_l'] / self.parameters['unit_t']
        self.conversion_factors["x-velocity"] = vel_u
        self.conversion_factors["y-velocity"] = vel_u
        self.conversion_factors["z-velocity"] = vel_u

    def _setup_nounits_units(self):
        for unit in mpc_conversion.keys():
            self.units[unit] = self.parameters['unit_l'] * mpc_conversion[unit] / mpc_conversion["cm"]
        for unit in sec_conversion.keys():
            self.time_units[unit] = self.parameters['unit_t'] / sec_conversion[unit]

    def _parse_parameter_file(self):
        # hardcoded for now
        # These should be explicitly obtained from the file, but for now that
        # will wait until a reorganization of the source tree and better
        # generalization.
        self.dimensionality = 3
        self.refine_by = 2
        self.parameters["HydroMethod"] = 'ramses'
        self.parameters["Time"] = 1. # default unit is 1...

        self.unique_identifier = \
            int(os.stat(self.parameter_filename)[stat.ST_CTIME])
        self.ramses_tree = _ramses_reader.RAMSES_tree_proxy(self.parameter_filename)
        rheader = self.ramses_tree.get_file_info()
        self.parameters.update(rheader)
        self.current_time = self.parameters['time'] * self.parameters['unit_t']
        self.domain_right_edge = np.ones(3, dtype='float64') \
                                           * rheader['boxlen']
        self.domain_left_edge = np.zeros(3, dtype='float64')
        self.domain_dimensions = np.ones(3, dtype='int32') * 2
        # This is likely not true, but I am not sure how to otherwise
        # distinguish them.
        mylog.warning("No current mechanism of distinguishing cosmological simulations in RAMSES!")
        self.cosmological_simulation = 1
        self.periodicity = (True, True, True)
        self.current_redshift = (1.0 / rheader["aexp"]) - 1.0
        self.omega_lambda = rheader["omega_l"]
        self.omega_matter = rheader["omega_m"]
        self.hubble_constant = rheader["H0"]

    @classmethod
    def _is_valid(self, *args, **kwargs):
        if not os.path.basename(args[0]).startswith("info_"): return False
        fn = args[0].replace("info_", "amr_").replace(".txt", ".out00001")
        print fn
        return os.path.exists(fn)




"""
This module is very simple.  It imports the configuration
we have written for yt.
Everything will be returned in a global config dictionary: ytcfg

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import ConfigParser, os, os.path, types

ytcfgDefaults = dict(
    serialize = 'True',
    onlydeserialize = 'False',
    timefunctions = 'False',
    logfile = 'False',
    coloredlogs = 'False',
    suppressstreamlogging = 'False',
    loglevel = '20',
    inline = 'False',
    numthreads = '-1',
    __withinreason = 'False',
    __withintesting = 'False',
    __parallel = 'False',
    __global_parallel_rank = '0',
    __global_parallel_size = '1',
    __topcomm_parallel_rank = '0',
    __topcomm_parallel_size = '1',
    __command_line = 'False',
    storeparameterfiles = 'False',
    parameterfilestore = 'parameter_files.csv',
    maximumstoredpfs = '500',
    loadfieldplugins = 'True',
    pluginfilename = 'my_plugins.py',
    parallel_traceback = 'False',
    pasteboard_repo = '',
    reconstruct_hierarchy = 'False',
    test_storage_dir = '/does/not/exist',
    test_data_dir = '/does/not/exist',
    enzo_db = '',
    hub_url = 'https://hub.yt-project.org/upload',
    hub_api_key = '',
    ipython_notebook = 'False',
    notebook_password = '',
    answer_testing_tolerance = '3',
    answer_testing_bitwise = 'False',
    gold_standard_filename = 'gold008',
    local_standard_filename = 'local001',
    sketchfab_api_key = 'None'
    )
# Here is the upgrade.  We're actually going to parse the file in its entirety
# here.  Then, if it has any of the Forbidden Sections, it will be rewritten
# without them.

__fn = os.path.expanduser("~/.yt/config")
if os.path.exists(__fn):
    f = open(__fn).read()
    if any(header in f for header in ["[lagos]","[raven]","[fido]","[enki]"]):
        print "***********************************************************"
        print "* Upgrading configuration file to new format; saving old. *"
        print "***********************************************************"
        # This is of the old format
        cp = ConfigParser.ConfigParser()
        cp.read(__fn)
        # NOTE: To avoid having the 'DEFAULT' section here,
        # we are not passing in ytcfgDefaults to the constructor.
        new_cp = ConfigParser.ConfigParser()
        new_cp.add_section("yt")
        for section in cp.sections():
            for option in cp.options(section):
                # We changed them all to lowercase
                if option.lower() in ytcfgDefaults:
                    new_cp.set("yt", option, cp.get(section, option))
                    print "Setting %s to %s" % (option, cp.get(section, option))
        open(__fn + ".old", "w").write(f)
        new_cp.write(open(__fn, "w"))
# Pathological check for Kraken
#elif os.path.exists("~/"):
#    if not os.path.exists("~/.yt"):
#            print "yt is creating a new directory, ~/.yt ."
#            os.mkdir(os.path.exists("~/.yt/"))
#    # Now we can read in and write out ...
#    new_cp = Configparser.ConfigParser(ytcfgDefaults)
#    new_cp.write(__fn)

class YTConfigParser(ConfigParser.ConfigParser):
    def __setitem__(self, key, val):
        self.set(key[0], key[1], val)

if os.path.exists(os.path.expanduser("~/.yt/config")):
    ytcfg = YTConfigParser(ytcfgDefaults)
    ytcfg.read(['yt.cfg', os.path.expanduser('~/.yt/config')])
else:
    ytcfg = YTConfigParser(ytcfgDefaults)
    ytcfg.read(['yt.cfg'])
if not ytcfg.has_section("yt"):
    ytcfg.add_section("yt")

# Now we have parsed the config file.  Overrides come from the command line.

# This should be implemented at some point.  The idea would be to have a set of
# command line options, fed through an option parser, that would override
# the settings in ytcfg.  *However*, because we want to have the command-line
# scripts work, we'd probably want to have them only be long options, and also
# along the lines of --yt-something=somethingelse.  The command line scripts
# would then not get their options from sys.argv, but instead from this module.



"""
A collection of helper functions, most generally for things
that SciPy doesn't have that I expected it to

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from yt.funcs import *
import yt.utilities.lib as lib

class UnilinearFieldInterpolator:
    def __init__(self, table, boundaries, field_names, truncate=False):
        r"""Initialize a 1D interpolator for field data.

        table : array
            The data table over which interpolation is performed.
        boundaries: tuple or array
            If a tuple, this should specify the upper and lower bounds 
            for the bins of the data table.  This assumes the bins are 
            evenly spaced.  If an array, this specifies the bins 
            explicitly.
        field_names: str
            Name of the field to be used as input data for interpolation.
        truncate : bool
            If False, an exception is raised if the input values are 
            outside the bounds of the table.  If True, extrapolation is 
            performed.
        
        Examples
        --------

        ad = pf.h.all_data()
        table_data = np.random.random(64)
        interp = UnilinearFieldInterpolator(table_data, (0.0, 1.0), "x",
                                            truncate=True)
        field_data = interp(ad)
        
        """
        self.table = table.astype('float64')
        self.truncate = truncate
        self.x_name = field_names
        if isinstance(boundaries, np.ndarray):
            if boundaries.size != table.shape[0]:
                mylog.error("Bins array not the same length as the data.")
                raise ValueError
            self.x_bins = boundaries
        else:
            x0, x1 = boundaries
            self.x_bins = np.linspace(x0, x1, table.shape[0]).astype('float64')
        
    def __call__(self, data_object):
        orig_shape = data_object[self.x_name].shape
        x_vals = data_object[self.x_name].ravel().astype('float64')

        x_i = (np.digitize(x_vals, self.x_bins) - 1).astype('int32')
        if np.any((x_i == -1) | (x_i == len(self.x_bins)-1)):
            if not self.truncate:
                mylog.error("Sorry, but your values are outside" + \
                            " the table!  Dunno what to do, so dying.")
                mylog.error("Error was in: %s", data_object)
                raise ValueError
            else:
                x_i = np.minimum(np.maximum(x_i,0), len(self.x_bins)-2)

        my_vals = np.zeros(x_vals.shape, dtype='float64')
        lib.UnilinearlyInterpolate(self.table, x_vals, self.x_bins, x_i, my_vals)
        return my_vals.reshape(orig_shape)

class BilinearFieldInterpolator:
    def __init__(self, table, boundaries, field_names, truncate=False):
        r"""Initialize a 2D interpolator for field data.

        table : array
            The data table over which interpolation is performed.
        boundaries: tuple
            Either a tuple of lower and upper bounds for the x and y bins 
            given as (x0, x1, y0, y1) or a tuple of two arrays containing the 
            x and y bins.
        field_names: list
            Names of the fields to be used as input data for interpolation.
        truncate : bool
            If False, an exception is raised if the input values are 
            outside the bounds of the table.  If True, extrapolation is 
            performed.
        
        Examples
        --------

        ad = pf.h.all_data()
        table_data = np.random.random((64, 64))
        interp = BilinearFieldInterpolator(table_data, (0.0, 1.0, 0.0, 1.0), 
                                           ["x", "y"],
                                           truncate=True)
        field_data = interp(ad)
        
        """
        self.table = table.astype('float64')
        self.truncate = truncate
        self.x_name, self.y_name = field_names
        if len(boundaries) == 4:
            x0, x1, y0, y1 = boundaries
            self.x_bins = np.linspace(x0, x1, table.shape[0]).astype('float64')
            self.y_bins = np.linspace(y0, y1, table.shape[1]).astype('float64')
        elif len(boundaries) == 2:
            if boundaries[0].size != table.shape[0]:
                mylog.error("X bins array not the same length as the data.")
                raise ValueError
            if boundaries[1].size != table.shape[1]:
                mylog.error("Y bins array not the same length as the data.")
                raise ValueError
            self.x_bins = boundaries[0]
            self.y_bins = boundaries[1]
        else:
            mylog.error("Boundaries must be given as (x0, x1, y0, y1) or as (x_bins, y_bins)")
            raise ValueError

    def __call__(self, data_object):
        orig_shape = data_object[self.x_name].shape
        x_vals = data_object[self.x_name].ravel().astype('float64')
        y_vals = data_object[self.y_name].ravel().astype('float64')

        x_i = (np.digitize(x_vals, self.x_bins) - 1).astype('int32')
        y_i = (np.digitize(y_vals, self.y_bins) - 1).astype('int32')
        if np.any((x_i == -1) | (x_i == len(self.x_bins)-1)) \
            or np.any((y_i == -1) | (y_i == len(self.y_bins)-1)):
            if not self.truncate:
                mylog.error("Sorry, but your values are outside" + \
                            " the table!  Dunno what to do, so dying.")
                mylog.error("Error was in: %s", data_object)
                raise ValueError
            else:
                x_i = np.minimum(np.maximum(x_i,0), len(self.x_bins)-2)
                y_i = np.minimum(np.maximum(y_i,0), len(self.y_bins)-2)

        my_vals = np.zeros(x_vals.shape, dtype='float64')
        lib.BilinearlyInterpolate(self.table,
                                 x_vals, y_vals, self.x_bins, self.y_bins,
                                 x_i, y_i, my_vals)
        return my_vals.reshape(orig_shape)

class TrilinearFieldInterpolator:
    def __init__(self, table, boundaries, field_names, truncate = False):
        r"""Initialize a 3D interpolator for field data.

        table : array
            The data table over which interpolation is performed.
        boundaries: tuple
            Either a tuple of lower and upper bounds for the x, y, and z bins 
            given as (x0, x1, y0, y1, z0, z1) or a tuple of three arrays 
            containing the x, y, and z bins.
        field_names: list
            Names of the fields to be used as input data for interpolation.
        truncate : bool
            If False, an exception is raised if the input values are 
            outside the bounds of the table.  If True, extrapolation is 
            performed.
        
        Examples
        --------

        ad = pf.h.all_data()
        table_data = np.random.random((64, 64, 64))
        interp = BilinearFieldInterpolator(table_data, 
                                           (0.0, 1.0, 0.0, 1.0, 0.0, 1.0), 
                                           ["x", "y", "z"],
                                           truncate=True)
        field_data = interp(ad)
        
        """
        self.table = table.astype('float64')
        self.truncate = truncate
        self.x_name, self.y_name, self.z_name = field_names
        if len(boundaries) == 6:
            x0, x1, y0, y1, z0, z1 = boundaries
            self.x_bins = np.linspace(x0, x1, table.shape[0]).astype('float64')
            self.y_bins = np.linspace(y0, y1, table.shape[1]).astype('float64')
            self.z_bins = np.linspace(z0, z1, table.shape[2]).astype('float64')
        elif len(boundaries) == 3:
            if boundaries[0].size != table.shape[0]:
                mylog.error("X bins array not the same length as the data.")
                raise ValueError
            if boundaries[1].size != table.shape[1]:
                mylog.error("Y bins array not the same length as the data.")
                raise ValueError
            if boundaries[2].size != table.shape[2]:
                mylog.error("Z bins array not the same length as the data.")
                raise ValueError
            self.x_bins = boundaries[0]
            self.y_bins = boundaries[1]
            self.z_bins = boundaries[2]
        else:
            mylog.error("Boundaries must be given as (x0, x1, y0, y1, z0, z1) or as (x_bins, y_bins, z_bins)")
            raise ValueError
        
    def __call__(self, data_object):
        orig_shape = data_object[self.x_name].shape
        x_vals = data_object[self.x_name].ravel().astype('float64')
        y_vals = data_object[self.y_name].ravel().astype('float64')
        z_vals = data_object[self.z_name].ravel().astype('float64')

        x_i = np.digitize(x_vals, self.x_bins) - 1
        y_i = np.digitize(y_vals, self.y_bins) - 1
        z_i = np.digitize(z_vals, self.z_bins) - 1
        if np.any((x_i == -1) | (x_i == len(self.x_bins)-1)) \
            or np.any((y_i == -1) | (y_i == len(self.y_bins)-1)) \
            or np.any((z_i == -1) | (z_i == len(self.z_bins)-1)):
            if not self.truncate:
                mylog.error("Sorry, but your values are outside" + \
                            " the table!  Dunno what to do, so dying.")
                mylog.error("Error was in: %s", data_object)
                raise ValueError
            else:
                x_i = np.minimum(np.maximum(x_i,0), len(self.x_bins)-2)
                y_i = np.minimum(np.maximum(y_i,0), len(self.y_bins)-2)
                z_i = np.minimum(np.maximum(z_i,0), len(self.z_bins)-2)

        my_vals = np.zeros(x_vals.shape, dtype='float64')
        lib.TrilinearlyInterpolate(self.table,
                                 x_vals, y_vals, z_vals,
                                 self.x_bins, self.y_bins, self.z_bins,
                                 x_i, y_i, z_i, my_vals)
        return my_vals.reshape(orig_shape)

        # Use notation from Paul Bourke's page on interpolation
        # http://local.wasp.uwa.edu.au/~pbourke/other/interpolation/
        x = (x_vals - self.x_bins[x_i]) / (self.x_bins[x_i+1] - self.x_bins[x_i])
        y = (y_vals - self.y_bins[y_i]) / (self.y_bins[y_i+1] - self.y_bins[y_i])
        z = (z_vals - self.z_bins[z_i]) / (self.z_bins[z_i+1] - self.z_bins[z_i])
        xm = (self.x_bins[x_i+1] - x_vals) / (self.x_bins[x_i+1] - self.x_bins[x_i])
        ym = (self.y_bins[y_i+1] - y_vals) / (self.y_bins[y_i+1] - self.y_bins[y_i])
        zm = (self.z_bins[z_i+1] - z_vals) / (self.z_bins[z_i+1] - self.z_bins[z_i])
        if np.any(np.isnan(self.table)):
            raise ValueError
        if np.any(np.isnan(x) | np.isnan(y) | np.isnan(z)):
            raise ValueError
        if np.any(np.isnan(xm) | np.isnan(ym) | np.isnan(zm)):
            raise ValueError
        my_vals  = self.table[x_i  ,y_i  ,z_i  ] * (xm*ym*zm)
        my_vals += self.table[x_i+1,y_i  ,z_i  ] * (x *ym*zm)
        my_vals += self.table[x_i  ,y_i+1,z_i  ] * (xm*y *zm)
        my_vals += self.table[x_i  ,y_i  ,z_i+1] * (xm*ym*z )
        my_vals += self.table[x_i+1,y_i  ,z_i+1] * (x *ym*z )
        my_vals += self.table[x_i  ,y_i+1,z_i+1] * (xm*y *z )
        my_vals += self.table[x_i+1,y_i+1,z_i  ] * (x *y *zm)
        my_vals += self.table[x_i+1,y_i+1,z_i+1] * (x *y *z )
        return my_vals.reshape(orig_shape)

def get_centers(pf, filename, center_cols, radius_col, unit='1'):
    """
    Return an iterator over EnzoSphere objects generated from the appropriate 
    columns in *filename*.  Optionally specify the *unit* radius is in.
    """
    sp_list = []
    for line in open(filename):
        if line.startswith("#"): continue
        vals = line.split()
        x,y,z = [float(vals[i]) for i in center_cols]
        r = float(vals[radius_col])
        yield pf.h.sphere([x,y,z], r/pf[unit])



"""
A light interface to libpng

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
cimport numpy as np
cimport cython
from libc.stdlib cimport malloc, realloc, free
from libc.string cimport memcpy
from cpython.string cimport PyString_FromStringAndSize
from fp_utils cimport imax, fmax, imin, fmin, iclip, fclip

from libc.stdio cimport fopen, fclose, FILE

cdef extern from "stdlib.h":
    # NOTE that size_t might not be int
    void *alloca(int)

# Idiom for accessing Python files, from Cython FAQ
# First, declare the Python macro to access files:
cdef extern from "Python.h":
    ctypedef struct FILE
    FILE* PyFile_AsFile(object)
    void  fprintf(FILE* f, char* s, char* s)

cdef extern from "png.h":
    ctypedef unsigned long png_uint_32
    ctypedef long png_int_32
    ctypedef unsigned short png_uint_16
    ctypedef short png_int_16
    ctypedef unsigned char png_byte
    ctypedef void            *png_voidp
    ctypedef png_byte        *png_bytep
    ctypedef png_uint_32     *png_uint_32p
    ctypedef png_int_32      *png_int_32p
    ctypedef png_uint_16     *png_uint_16p
    ctypedef png_int_16      *png_int_16p
    ctypedef char            *png_charp
    ctypedef char            *png_const_charp
    ctypedef FILE            *png_FILE_p

    ctypedef struct png_struct:
        png_voidp io_ptr
    ctypedef png_struct      *png_structp

    ctypedef struct png_info:
        pass
    ctypedef png_info        *png_infop

    ctypedef struct png_color_8:
        png_byte red
        png_byte green
        png_byte blue
        png_byte gray
        png_byte alpha
    ctypedef png_color_8 *png_color_8p

    cdef png_const_charp PNG_LIBPNG_VER_STRING

    # Note that we don't support error or warning functions
    png_structp png_create_write_struct(
        png_const_charp user_png_ver, png_voidp error_ptr,
        void *error_fn, void *warn_fn)
    
    png_infop png_create_info_struct(png_structp png_ptr)

    void png_init_io(png_structp png_ptr, png_FILE_p fp)

    void png_set_IHDR(png_structp png_ptr, png_infop info_ptr,
        png_uint_32 width, png_uint_32 height, int bit_depth,
        int color_type, int interlace_method, int compression_method,
        int filter_method)

    cdef int PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE
    cdef int PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_ADAM7
    cdef int PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE

    void png_set_pHYs(png_structp png_ptr, png_infop info_ptr,
        png_uint_32 res_x, png_uint_32 res_y, int unit_type)

    cdef int PNG_RESOLUTION_METER

    void png_set_sBIT(png_structp png_ptr, png_infop info_ptr,
        png_color_8p sig_bit)

    ctypedef void (*png_rw_ptr) (png_structp, png_bytep, size_t)
    ctypedef void (*png_flush_ptr) (png_structp)
    void png_set_write_fn(png_structp png_ptr, png_voidp io_ptr,
                          png_rw_ptr write_data_fn,
                          png_flush_ptr output_flush_fn)
    png_voidp png_get_io_ptr (png_structp png_ptr)

    void png_write_info(png_structp png_ptr, png_infop info_ptr)
    void png_set_rows(png_structp png_ptr, png_infop info_ptr,
                      png_bytep *row_pointers)
    void png_write_image(png_structp png_ptr, png_bytep *image)
    void png_write_end(png_structp png_ptr, png_infop info_ptr)
    void png_write_png(png_structp png_ptr, png_infop info_ptr,
                       int transforms, png_voidp params)
    cdef int PNG_TRANSFORM_IDENTITY

    void png_destroy_write_struct(
        png_structp *png_ptr_ptr, png_infop *info_ptr_ptr)

def write_png_to_file(np.ndarray[np.uint8_t, ndim=3] buffer,
                      object py_fileobj, int dpi=100,
                      int close = 0):

    # This is something of a translation of the matplotlib _png module
    cdef png_byte *pix_buffer = <png_byte *> buffer.data
    cdef int width = buffer.shape[1]
    cdef int height = buffer.shape[0]
    cdef FILE *fileobj = PyFile_AsFile(py_fileobj)

    cdef png_bytep *row_pointers
    cdef png_structp png_ptr
    cdef png_infop info_ptr

    cdef png_color_8 sig_bit
    cdef png_uint_32 row

    row_pointers = <png_bytep *> alloca(sizeof(png_bytep) * height)

    for row in range(height):
        row_pointers[row] = pix_buffer + row * width * 4
    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)
    info_ptr = png_create_info_struct(png_ptr)
    
    # Um we are ignoring setjmp sorry guys

    png_init_io(png_ptr, fileobj)

    png_set_IHDR(png_ptr, info_ptr, width, height, 8,
                 PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE,
                 PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE)

    cdef size_t dots_per_meter = <size_t> (dpi / (2.54 / 100.0))
    png_set_pHYs(png_ptr, info_ptr, dots_per_meter, dots_per_meter,
                 PNG_RESOLUTION_METER)

    sig_bit.gray = 0
    sig_bit.red = sig_bit.green = sig_bit.blue = sig_bit.alpha = 8

    png_set_sBIT(png_ptr, info_ptr, &sig_bit)

    png_write_info(png_ptr, info_ptr)
    png_write_image(png_ptr, row_pointers)
    png_write_end(png_ptr, info_ptr)

    if close == 1: fclose(fileobj)
    png_destroy_write_struct(&png_ptr, &info_ptr)

def write_png(np.ndarray[np.uint8_t, ndim=3] buffer,
              char *filename, int dpi=100):

    # This is something of a translation of the matplotlib _png module
    cdef png_byte *pix_buffer = <png_byte *> buffer.data
    cdef int width = buffer.shape[1]
    cdef int height = buffer.shape[0]

    cdef FILE* fileobj = fopen(filename, "wb")
    cdef png_bytep *row_pointers
    cdef png_structp png_ptr
    cdef png_infop info_ptr

    cdef png_color_8 sig_bit
    cdef png_uint_32 row

    row_pointers = <png_bytep *> alloca(sizeof(png_bytep) * height)

    for row in range(height):
        row_pointers[row] = pix_buffer + row * width * 4
    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)
    info_ptr = png_create_info_struct(png_ptr)
    
    # Um we are ignoring setjmp sorry guys

    png_init_io(png_ptr, fileobj)

    png_set_IHDR(png_ptr, info_ptr, width, height, 8,
                 PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE,
                 PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE)

    cdef size_t dots_per_meter = <size_t> (dpi / (2.54 / 100.0))
    png_set_pHYs(png_ptr, info_ptr, dots_per_meter, dots_per_meter,
                 PNG_RESOLUTION_METER)

    sig_bit.gray = 0
    sig_bit.red = sig_bit.green = sig_bit.blue = sig_bit.alpha = 8

    png_set_sBIT(png_ptr, info_ptr, &sig_bit)

    png_write_info(png_ptr, info_ptr)
    png_write_image(png_ptr, row_pointers)
    png_write_end(png_ptr, info_ptr)

    fclose(fileobj)
    png_destroy_write_struct(&png_ptr, &info_ptr)


# Much of this is inspired by and translated from this StackOverflow question:
# http://stackoverflow.com/questions/1821806/how-to-encode-png-to-buffer-using-libpng

cdef public struct mem_encode:
    char *buffer
    size_t size

cdef public void my_png_write_data(png_structp png_ptr, png_bytep data,
                                   size_t length):
    cdef png_voidp temp = png_get_io_ptr(png_ptr)
    cdef mem_encode *p = <mem_encode *> temp
    cdef size_t nsize = p.size + length
    if p.buffer != NULL:
        p.buffer = <char *> realloc(p.buffer, nsize)
    else:
        p.buffer = <char *> malloc(nsize)
    memcpy(p.buffer + p.size, data, length)
    p.size += length

cdef public void my_png_flush(png_structp png_ptr):
    return

def write_png_to_string(np.ndarray[np.uint8_t, ndim=3] buffer, int dpi=100,
                        int gray = 0):

    # This is something of a translation of the matplotlib _png module
    cdef png_byte *pix_buffer = <png_byte *> buffer.data
    cdef int width = buffer.shape[1]
    cdef int height = buffer.shape[0]

    cdef png_bytep *row_pointers
    cdef png_structp png_ptr
    cdef png_infop info_ptr

    cdef png_color_8 sig_bit
    cdef png_uint_32 row

    row_pointers = <png_bytep *> alloca(sizeof(png_bytep) * height)

    for row in range(height):
        row_pointers[row] = pix_buffer + row * width * 4
    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)
    info_ptr = png_create_info_struct(png_ptr)
    
    # Um we are ignoring setjmp sorry guys

    cdef int im_type, interlace_type
    interlace_type = PNG_INTERLACE_NONE
    if gray == 0:
        im_type = PNG_COLOR_TYPE_RGB_ALPHA
        sig_bit.gray = 0
        sig_bit.red = sig_bit.green = sig_bit.blue = sig_bit.alpha = 8
    elif gray == 1:
        im_type = PNG_COLOR_TYPE_GRAY
        sig_bit.gray = 8
        sig_bit.red = sig_bit.green = sig_bit.blue = sig_bit.alpha = 0
    else:
        raise RuntimeError

    png_set_IHDR(png_ptr, info_ptr, width, height, 8,
                 im_type, interlace_type,
                 PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE)

    cdef size_t dots_per_meter = <size_t> (dpi / (2.54 / 100.0))
    png_set_pHYs(png_ptr, info_ptr, dots_per_meter, dots_per_meter,
                 PNG_RESOLUTION_METER)


    png_set_sBIT(png_ptr, info_ptr, &sig_bit)

    cdef mem_encode state 
    state.buffer = NULL
    state.size = 0

    png_set_write_fn(png_ptr, <png_voidp> &state, my_png_write_data, NULL)
    png_set_rows(png_ptr, info_ptr, row_pointers)
    png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL)

    png_destroy_write_struct(&png_ptr, &info_ptr)

    pp = PyString_FromStringAndSize(state.buffer, state.size)
    if state.buffer != NULL: free(state.buffer)
    return pp

def add_points_to_image(
        np.ndarray[np.uint8_t, ndim=3] buffer,
        np.ndarray[np.float64_t, ndim=1] px,
        np.ndarray[np.float64_t, ndim=1] py,
        np.float64_t pv):
    cdef int i, j, k, pi
    cdef int np = px.shape[0]
    cdef int xs = buffer.shape[0]
    cdef int ys = buffer.shape[1]
    cdef int v
    v = iclip(<int>(pv * 255), 0, 255)
    for pi in range(np):
        j = <int> (xs * px[pi])
        i = <int> (ys * py[pi])
        for k in range(3):
            buffer[i, j, k] = 0
    return
    #for i in range(xs):
    #    for j in range(ys):
    #        for k in range(3):
    #            v = buffer[i, j, k]
    #            buffer[i, j, k] = iclip(v, 0, 255)



"""
YT is a package written primarily in Python designed to make the task of
running Enzo easier.  It contains facilities for creating Enzo data (currently
in prototype form) as well as runnning Enzo simulations, simulating the actions
of Enzo on various existing data, and analyzing output from Enzo in a
wide-variety of methods.

An ever-growing collection of documentation is also available at
http://yt-project.org/doc/ . Additionally, there is a
project site at http://yt-project.org/ with recipes, a wiki, a variety of
ways of peering into the version control, and a bug-reporting system.

YT is divided into several packages.

frontends
---------

This is where interfaces to codes are created.  Within each subdirectory of
yt/frontends/ there must exist the following files, even if empty:

* data_structures.py, where subclasses of AMRGridPatch, StaticOutput and
  AMRHierarchy are defined.
* io.py, where a subclass of IOHandler is defined.
* misc.py, where any miscellaneous functions or classes are defined.
* definitions.py, where any definitions specific to the frontend are
  defined.  (i.e., header formats, etc.)

visualization
-------------

This is where all visualization modules are stored.  This includes plot
collections, the volume rendering interface, and pixelization frontends.

data_objects
------------

All objects that handle data, processed or unprocessed, not explicitly
defined as visualization are located in here.  This includes the base
classes for data regions, covering grids, time series, and so on.  This
also includes derived fields and derived quantities.

analysis_modules
----------------

This is where all mechanisms for processing data live.  This includes
things like clump finding, halo profiling, halo finding, and so on.  This
is something of a catchall, but it serves as a level of greater
abstraction that simply data selection and modification.

gui
---

This is where all GUI components go.  Typically this will be some small
tool used for one or two things, which contains a launching mechanism on
the command line.

utilities
---------

All broadly useful code that doesn't clearly fit in one of the other
categories goes here.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__version__ = "2.5-dev"

def run_nose(verbose=False, run_answer_tests=False, answer_big_data=False):
    import nose, os, sys
    from yt.config import ytcfg
    nose_argv = sys.argv
    nose_argv += ['--exclude=answer_testing','--detailed-errors']
    if verbose:
        nose_argv.append('-v')
    if run_answer_tests:
        nose_argv.append('--with-answer-testing')
    if answer_big_data:
        nose_argv.append('--answer-big-data')
    log_suppress = ytcfg.getboolean("yt","suppressStreamLogging")
    ytcfg["yt","suppressStreamLogging"] = 'True'
    initial_dir = os.getcwd()
    yt_file = os.path.abspath(__file__)
    yt_dir = os.path.dirname(yt_file)
    os.chdir(yt_dir)
    try:
        nose.run(argv=nose_argv)
    finally:
        os.chdir(initial_dir)
        ytcfg["yt","suppressStreamLogging"] = log_suppress



"""
Castro-specific fields

Author: J. S. Oishi <jsoishi@gmail.com>
Affiliation: UC Berkeley
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2010 J. S. Oishi, Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from yt.data_objects.field_info_container import \
    FieldInfoContainer, \
    FieldInfo, \
    NullFunc, \
    TranslationFunc, \
    ValidateParameter, \
    ValidateDataField, \
    ValidateProperty, \
    ValidateSpatial, \
    ValidateGridType
import yt.data_objects.universal_fields
from yt.utilities.physical_constants import mh, kboltz

translation_dict = {
    "x-velocity": "xvel",
    "y-velocity": "yvel",
    "z-velocity": "zvel",
    "Density": "density",
    "Total_Energy": "eden",
    "Temperature": "temperature",
    "x-momentum": "xmom",
    "y-momentum": "ymom",
    "z-momentum": "zmom"
}

# Setup containers for fields possibly in the output files
KnownCastroFields = FieldInfoContainer()
add_castro_field = KnownCastroFields.add_field

# and always derived ones
CastroFieldInfo = FieldInfoContainer.create_with_fallback(FieldInfo)
add_field = CastroFieldInfo.add_field

# Start adding fields
add_castro_field("density", function=NullFunc, take_log=True,
                 units=r"\rm{g}/\rm{cm}^3")

# fix projected units
KnownCastroFields["density"]._projected_units = r"\rm{g}/\rm{cm}^2"

add_castro_field("eden", function=NullFunc, take_log=True,
                 validators = [ValidateDataField("eden")],
                 units=r"\rm{erg}/\rm{cm}^3")

add_castro_field("xmom", function=NullFunc, take_log=False,
                 validators = [ValidateDataField("xmom")],
                 units=r"\rm{g}/\rm{cm^2\ s}")

add_castro_field("ymom", function=NullFunc, take_log=False,
                 validators = [ValidateDataField("ymom")],
                 units=r"\rm{gm}/\rm{cm^2\ s}")

add_castro_field("zmom", function=NullFunc, take_log=False,
                 validators = [ValidateDataField("zmom")],
                 units=r"\rm{g}/\rm{cm^2\ s}")

# Now populate derived fields
for mine, theirs in translation_dict.items():
    if KnownCastroFields.has_key(theirs):
        add_field(theirs, function=TranslationFunc(mine),
                  take_log=KnownCastroFields[theirs].take_log)

# Now fallbacks, in case these fields are not output
def _xVelocity(field, data):
    """ Generate x-velocity from x-momentum and density. """
    return data["xmom"] / data["density"]

add_field("x-velocity", function=_xVelocity, take_log=False,
          units=r'\rm{cm}/\rm{s}')

def _yVelocity(field, data):
    """ Generate y-velocity from y-momentum and density. """
    return data["ymom"] / data["density"]

add_field("y-velocity", function=_yVelocity, take_log=False,
          units=r'\rm{cm}/\rm{s}')

def _zVelocity(field, data):
    """ Generate z-velocity from z-momentum and density. """
    return data["zmom"] / data["density"]

add_field("z-velocity", function=_zVelocity, take_log=False,
          units=r'\rm{cm}/\rm{s}')

def _ThermalEnergy(field, data):
    """
    Generate thermal (gas energy). Dual Energy Formalism was implemented by
    Stella, but this isn't how it's called, so I'll leave that commented out for
    now.

    """
    #if data.pf["DualEnergyFormalism"]:
    #    return data["Gas_Energy"]
    #else:
    return data["Total_Energy"] - 0.5 * data["density"] * (
        data["x-velocity"]**2.0
        + data["y-velocity"]**2.0
        + data["z-velocity"]**2.0 )

add_field("ThermalEnergy", function=_ThermalEnergy,
          units=r"\rm{ergs}/\rm{cm^3}")

def _Pressure(field, data):
    """
    M{(Gamma-1.0)*e, where e is thermal energy density
    
    NB: this will need to be modified for radiation

    """
    return (data.pf["Gamma"] - 1.0) * data["ThermalEnergy"]

add_field("Pressure", function=_Pressure, units=r"\rm{dyne}/\rm{cm}^{2}")

def _Temperature(field, data):
    return ((data.pf["Gamma"] - 1.0) * data.pf["mu"] * mh *
            data["ThermalEnergy"] / (kboltz * data["Density"]))

add_field("Temperature", function=_Temperature, units=r"\rm{Kelvin}",
          take_log=False)

def _convertParticleMassMsun(data):
    return 1.0 / 1.989e33
def _ParticleMassMsun(field, data):
    return data["particle_mass"]

add_field("ParticleMassMsun",
          function=_ParticleMassMsun, validators=[ValidateSpatial(0)],
          particle_type=True, convert_function=_convertParticleMassMsun,
          particle_convert_function=_ParticleMassMsun)

# Fundamental fields that are usually/always output:
#   density
#   xmom
#   ymom
#   zmom
#   rho_E
#   rho_e
#   Temp
#
# "Derived" fields that are sometimes output:
#   x_velocity
#   y_velocity
#   z_velocity
#   magvel
#   grav_x
#   grav_y
#   grav_z
#   maggrav
#   magvort
#   pressure
#   entropy
#   divu
#   eint_e (e as derived from the "rho e" variable)
#   eint_E (e as derived from the "rho E" variable)



"""
A read-eval-print-loop that is served up through Bottle and accepts its
commands through HTTP

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import json
import os

from .bottle_mods import preroute
from .basic_repl import ProgrammaticREPL

local_dir = os.path.dirname(__file__)

class HTTPREPL(ProgrammaticREPL):

    def __init__(self, locals=None):
        # First we do the standard initialization
        ProgrammaticREPL.__init__(self, locals)
        # Now, since we want to only preroute functions we know about, and
        # since they have different arguments, and most of all because we only
        # want to add them to the routing tables (which are a singleton for the
        # entire interpreter state) we apply all the pre-routing now, rather
        # than through metaclasses or other fancy decorating.
        preroute_table = dict(index = ("/", "GET"),
                              push = ("/push", "POST"),
                              dir = ("/dir", "GET"),
                              doc = ("/doc", "GET"),
                              resources = ("/resources/:val", "GET"))
        for v, args in preroute_table.items():
            preroute(args[0], method=args[1])(getattr(self, v))

    def index(self):
        """Return an HTTP-based Read-Eval-Print-Loop terminal."""
        # For now this doesn't work!  We will need to move to a better method
        # for this.
        vals = open(os.path.join(local_dir, "httprepl.html")).read()
        return vals
        
    def push(self):
        """Push 'line' and return exec results as a bare response."""
        line = request.POST['line']
        result = ProgrammaticREPL.push(self, line)
        new_values = self.locals.pop("new_values", "")
        if result is None:
            # More input lines needed.
            response.status = 204
        return json.dumps( dict(text = result, new_values = new_values ))

    def dir(self):
        """Push 'line' and return result of eval on the final expr."""
        line = request.GET['line']
        result = ProgrammaticREPL.dir(self, line)
        if not result:
            response.status = 204
            return
        return repr(result)

    def doc(self):
        """Push 'line' and return result of getargspec on the final expr."""
        line = request.GET['line']
        result = ProgrammaticREPL.doc(self, line)
        if not result:
            response.status = 204
        return result

    def resources(self, val):
        pp = os.path.join(local_dir, "resources", val)
        if not os.path.exists(pp):
            response.status = 404
            return
        return open(pp).read()




#!/usr/bin/env python
import setuptools


def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('analysis_modules', parent_package, top_path)
    config.make_config_py()  # installs __config__.py
    #config.make_svn_version_py()
    config.add_subpackage("absorption_spectrum")
    config.add_subpackage("coordinate_transformation")
    config.add_subpackage("cosmological_observation")
    config.add_subpackage("halo_finding")
    config.add_subpackage("halo_mass_function")
    config.add_subpackage("halo_merger_tree")
    config.add_subpackage("halo_profiler")
    config.add_subpackage("hierarchy_subset")
    config.add_subpackage("level_sets")
    config.add_subpackage("radial_column_density")
    config.add_subpackage("spectral_integrator")
    config.add_subpackage("star_analysis")
    config.add_subpackage("two_point_functions")
    config.add_subpackage("radmc3d_export")
    return config



#!/usr/bin/env python
import setuptools
import os
import sys


def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('yt', parent_package, top_path)
    config.add_subpackage('analysis_modules')
    config.add_subpackage('data_objects')
    config.add_subpackage('frontends')
    config.add_subpackage('gui')
    config.add_subpackage('utilities')
    config.add_subpackage('visualization')
    config.make_config_py()
    #config.make_svn_version_py()
    return config



"""
TIGER-specific data structures

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.funcs import *
from yt.data_objects.grid_patch import \
           AMRGridPatch
from yt.data_objects.hierarchy import \
           AMRHierarchy
from yt.data_objects.static_output import \
           StaticOutput

from yt.data_objects.field_info_container import \
    FieldInfoContainer, NullFunc
from .fields import TigerFieldInfo, KnownTigerFields

class TigerGrid(AMRGridPatch):
    _id_offset = 0

    def __init__(self, id, hierarchy, left_edge, right_edge, left_dims, right_dims):
        AMRGridPatch.__init__(self, id, hierarchy = hierarchy)
        self.LeftEdge = left_edge
        self.RightEdge = right_edge
        self.Level = 0
        self.NumberOfParticles = 0
        self.left_dims = np.array(left_dims, dtype='int32')
        self.right_dims = np.array(right_dims, dtype='int32')
        self.ActiveDimensions = self.right_dims - self.left_dims
        self.Parent = None
        self.Children = []

    @property
    def child_mask(self):
        return np.ones(self.ActiveDimensions, dtype='int32')

    def __repr__(self):
        return "TigerGrid_%04i (%s)" % (self.id, self.ActiveDimensions)

class TigerHierarchy(AMRHierarchy):

    grid = TigerGrid

    def __init__(self, pf, data_style):
        self.directory = pf.fullpath
        self.data_style = data_style
        AMRHierarchy.__init__(self, pf, data_style)

    def _count_grids(self):
        # Tiger is unigrid
        self.ngdims = [i/j for i,j in
                izip(self.pf.root_size, self.pf.max_grid_size)]
        self.num_grids = np.prod(self.ngdims)
        self.max_level = 0

    def _setup_classes(self):
        dd = self._get_data_reader_dict()
        AMRHierarchy._setup_classes(self, dd)
        self.object_types.sort()

    def _parse_hierarchy(self):
        grids = []
        # We need to fill in dims, LE, RE, level, count
        dims, LE, RE, levels, counts = [], [], [], [], []
        DLE = self.pf.domain_left_edge
        DRE = self.pf.domain_right_edge 
        DW = DRE - DLE
        gds = DW / self.ngdims
        rd = [self.pf.root_size[i]-self.pf.max_grid_size[i] for i in range(3)]
        glx, gly, glz = np.mgrid[DLE[0]:DRE[0]-gds[0]:self.ngdims[0]*1j,
                                 DLE[1]:DRE[1]-gds[1]:self.ngdims[1]*1j,
                                 DLE[2]:DRE[2]-gds[2]:self.ngdims[2]*1j]
        gdx, gdy, gdz = np.mgrid[0:rd[0]:self.ngdims[0]*1j,
                                 0:rd[1]:self.ngdims[1]*1j,
                                 0:rd[2]:self.ngdims[2]*1j]
        LE, RE, levels, counts = [], [], [], []
        i = 0
        for glei, gldi in izip(izip(glx.flat, gly.flat, glz.flat),
                               izip(gdx.flat, gdy.flat, gdz.flat)):
            gld = np.array(gldi)
            gle = np.array(glei)
            gre = gle + gds
            g = self.grid(i, self, gle, gre, gld, gld+self.pf.max_grid_size)
            grids.append(g)
            dims.append(self.pf.max_grid_size)
            LE.append(g.LeftEdge)
            RE.append(g.RightEdge)
            levels.append(g.Level)
            counts.append(g.NumberOfParticles)
            i += 1
        self.grids = np.empty(len(grids), dtype='object')
        for gi, g in enumerate(grids): self.grids[gi] = g
        self.grid_dimensions[:] = np.array(dims, dtype='int64')
        self.grid_left_edge[:] = np.array(LE, dtype='float64')
        self.grid_right_edge[:] = np.array(RE, dtype='float64')
        self.grid_levels.flat[:] = np.array(levels, dtype='int32')
        self.grid_particle_count.flat[:] = np.array(counts, dtype='int32')

    def _populate_grid_objects(self):
        # We don't need to do anything here
        for g in self.grids: g._setup_dx()

    def _detect_fields(self):
        self.file_mapping = {"Density" : "rhob",
                             "Temperature" : "temp"}

    @property
    def field_list(self):
        return self.file_mapping.keys()

    def _setup_derived_fields(self):
        self.derived_field_list = []

class TigerStaticOutput(StaticOutput):
    _hierarchy_class = TigerHierarchy
    _fieldinfo_fallback = TigerFieldInfo
    _fieldinfo_known = KnownTigerFields

    def __init__(self, rhobname, root_size, max_grid_size=128,
                 data_style='tiger', storage_filename = None):
        StaticOutput.__init__(self, rhobname, data_style)
        self.storage_filename = storage_filename
        self.basename = rhobname[:-4]
        if not os.path.exists(self.basename + "rhob"):
            print "%s doesn't exist, don't know how to handle this!" % (
                        self.basename + "rhob")
            raise IOError
        if not iterable(root_size): root_size = (root_size,) * 3
        self.root_size = root_size
        if not iterable(max_grid_size): max_grid_size = (max_grid_size,) * 3
        self.max_grid_size = max_grid_size

        self.field_info = FieldInfoContainer.create_with_fallback(
                            self._fieldinfo_fallback)

        # We assume that we have basename + "rhob" and basename + "temp"
        # to get at our various parameters.

        # First we get our our header:
        
        header = [
            ('i', 'dummy0'),
            ('f', 'ZR'),
            ('f', 'OMEGA0'),
            ('f', 'FLAM0'),
            ('f', 'OMEGAB'),
            ('f', 'H0'),
            ('f', 'BOXL0'),
            ('i', 'dummy1'),
            ]

        h_fmt, h_key = zip(*header)
        header_string = "".join(h_fmt)

        fs = open(self.basename + "rhob")
        header_raw = read_struct(fs, header_string)
        self.parameters.update(dict(zip(h_key, header_raw)))

        if "InitialTime" not in self.parameters:
            self.current_time = 0.0
        self.unique_identifier = \
            int(os.stat(self.parameter_filename)[ST_CTIME])
        self.parameters['TopGridDimensions'] = root_size
        self.parameters['TopGridRank'] = 3
        self.units["Density"] = 1.0
        self.parameters['RefineBy'] = 2

    def _set_units(self):
        self.domain_left_edge = np.zeros(3, dtype='float64')
        self.domain_right_edge = np.ones(3, dtype='float64')
        self.units = {}
        self.time_units = {}
        self.time_units['1'] = 1
        self.units['1'] = 1.0
        self.units['cm'] = 1.0 # This is just plain false
        self.units['unitary'] = 1.0 / (self["DomainRightEdge"] - self["DomainLeftEdge"]).max()

    def _parse_parameter_file(self):
        pass

    @classmethod
    def _is_valid(self, *args, **kwargs):
        return os.path.exists(args[0] + "rhob")





"""
RAMSES-specific IO

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from collections import defaultdict
import numpy as np

from yt.utilities.io_handler import \
    BaseIOHandler
from yt.utilities.logger import ytLogger as mylog

class IOHandlerRAMSES(BaseIOHandler):
    _data_style = "ramses"

    def __init__(self, ramses_tree, *args, **kwargs):
        self.ramses_tree = ramses_tree
        BaseIOHandler.__init__(self, *args, **kwargs)

    def _read_data(self, grid, field):
        tr = np.zeros(grid.ActiveDimensions, dtype='float64')
        filled = np.zeros(grid.ActiveDimensions, dtype='int32')
        to_fill = grid.ActiveDimensions.prod()
        grids = [grid]
        l_delta = 0
        varindex = self.ramses_tree.field_ind[field]
        while to_fill > 0 and len(grids) > 0:
            next_grids = []
            for g in grids:
                to_fill -= self.ramses_tree.read_grid(varindex, field,
                        grid.get_global_startindex(), grid.ActiveDimensions,
                        tr, filled, g.Level, 2**l_delta, g.locations)
                next_grids += g.Parent
            grids = next_grids
            l_delta += 1
        return tr

    def preload(self, grids, sets):
        if len(grids) == 0: return
        domain_keys = defaultdict(list)
        pf_field_list = grids[0].pf.h.field_list
        sets = [dset for dset in list(sets) if dset in pf_field_list]
        exc = self._read_exception
        for g in grids:
            domain_keys[g.domain].append(g)
        for domain, grids in domain_keys.items():
            mylog.debug("Starting read of domain %s (%s)", domain, sets)
            for field in sets:
                for g in grids:
                    self.queue[g.id][field] = self._read_data(g, field)
                print "Clearing", field, domain
                self.ramses_tree.clear_tree(field, domain - 1)
        mylog.debug("Finished read of %s", sets)

    def modify(self, data): return data



#!/usr/bin/env python
import setuptools
import os
import sys
import os.path
import glob


def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('ramses', parent_package, top_path)
    config.add_extension("_ramses_reader",
        ["yt/frontends/ramses/_ramses_reader.pyx"],
        language="c++",
        include_dirs=["yt/frontends/ramses/ramses_headers/"],
        libraries=["stdc++"],
        depends=glob.glob("yt/frontends/ramses/ramses_headers/*.hh")
        )
    config.make_config_py()  # installs __config__.py
    #config.make_svn_version_py()
    return config



"""
Data structures for Nyx.

Author: Casey W. Stark <caseywstark@gmail.com>
Affiliation: UC Berkeley
Author: J. S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Casey W. Stark, J. S. Oishi, Matthew Turk.  All Rights
  Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from collections import defaultdict
import itertools
import os
import re
from stat import ST_CTIME
from string import strip, rstrip
import weakref

import numpy as np

from yt.funcs import *
from yt.data_objects.grid_patch import AMRGridPatch
from yt.data_objects.hierarchy import AMRHierarchy
from yt.data_objects.static_output import StaticOutput
from yt.data_objects.field_info_container import \
    FieldInfoContainer, NullFunc
from yt.utilities.lib import get_box_grids_level
from yt.utilities.definitions import \
    mpc_conversion, sec_conversion

from .definitions import parameter_type_dict, nyx_to_enzo_dict, \
                         fab_header_pattern, nyx_particle_field_names
from .utils import boxlib_bool_to_int
from .fields import NyxFieldInfo, add_nyx_field, KnownNyxFields


class NyxGrid(AMRGridPatch):
    _id_offset = 0

    def __init__(self, left_edge, right_edge, index, level, filename, offset,
                 dimensions, start, stop, **kwargs):
        """ Build a grid that understands Nyx's boxlib format. """
        # passes index as the patch ``id``
        AMRGridPatch.__init__(self, index, **kwargs)
        self.filename = filename
        self._offset = offset

        # @todo: enzo-isms.
        # Must copy to avoid refs, in case the user alters the args later.
        self.ActiveDimensions = (dimensions.copy()).astype('int32')
        self.start_index = start.copy()
        self.stop_index = stop.copy()
        self.LeftEdge  = left_edge.copy()
        self.RightEdge = right_edge.copy()
        self.index = index
        self.Level = level

    def get_global_startindex(self):
        return self.start_index

    def _prepare_grid(self):
        """ Copies all the appropriate attributes from the hierarchy. """
        h = self.hierarchy  # alias
        h.grid_levels[self.id, 0] = self.Level
        h.grid_left_edge[self.id,:] = self.LeftEdge[:]
        h.grid_right_edge[self.id,:] = self.RightEdge[:]

        # Might still work
        #self.Time = h.gridTimes[self.id,0]
        #self.NumberOfParticles = h.gridNumberOfParticles[self.id,0]

        # @todo: enzo-isms
        self.field_indexes = h.field_indexes
        self.Children = h.gridTree[self.id]
        pIDs = h.gridReverseTree[self.id]

        if len(pIDs) > 0:
            self.Parent = [weakref.proxy(h.grids[pID]) for pID in pIDs]
        else:
            # must be root grid
            self.Parent = None

    def _setup_dx(self):
        # has already been read in and stored in hierarchy
        dx = self.hierarchy.grid_dxs[self.index][0]
        dy = self.hierarchy.grid_dys[self.index][0]
        dz = self.hierarchy.grid_dzs[self.index][0]
        self.dds = np.array([dx, dy, dz])
        self.field_data['dx'], self.field_data['dy'], self.field_data['dz'] = self.dds

    def __repr__(self):
        return "NyxGrid_%04i" % (self.id)

class NyxHierarchy(AMRHierarchy):
    grid = NyxGrid

    def __init__(self, pf, data_style="nyx_native"):
        self.field_indexes = {}
        self.parameter_file = weakref.proxy(pf)
        self.directory = pf.path
        header_path = os.path.join(self.directory, "Header")  # make a kwarg?

        self.data_style = data_style
        #self._setup_classes()

        # This also sets up the grid objects
        self.read_global_header(header_path)
        self.read_particle_header()
        self.__cache_endianness(self.levels[-1].grids[-1])

        AMRHierarchy.__init__(self, pf, self.data_style)
        self._setup_data_io()
        self._setup_field_list()
        self._populate_hierarchy()

    def read_global_header(self, header_path):
        """ Read the global header file for an Nyx plotfile output. """
        counter = 0
        header_file = open(header_path, 'r')
        self._global_header_lines = header_file.readlines()

        # parse the file
        self.nyx_pf_version = self._global_header_lines[0].rstrip()
        self.n_fields = int(self._global_header_lines[1])

        # why the 2?
        counter = self.n_fields + 2
        self.field_list = []
        for i, line in enumerate(self._global_header_lines[2:counter]):
            self.field_list.append(line.rstrip())

        # figure out dimensions and make sure it's 3D
        self.dimension = int(self._global_header_lines[counter])
        if self.dimension != 3:
            raise RunTimeError("Current data is %iD. yt only supports Nyx data in 3D" % self.dimension)

        counter += 1
        self.Time = float(self._global_header_lines[counter])
        counter += 1
        self.finest_grid_level = int(self._global_header_lines[counter])
        self.n_levels = self.finest_grid_level + 1
        counter += 1

        # quantities with _unnecessary are also stored in the inputs
        # file and are not needed.  they are read in and stored in
        # case in the future we want to enable a "backwards" way of
        # taking the data out of the Header file and using it to fill
        # in in the case of a missing inputs file
        self.domainLeftEdge_unnecessary = np.array(map(float, self._global_header_lines[counter].split()))
        counter += 1
        self.domainRightEdge_unnecessary = np.array(map(float, self._global_header_lines[counter].split()))
        counter += 1
        self.refinementFactor_unnecessary = self._global_header_lines[counter].split() #np.array(map(int, self._global_header_lines[counter].split()))
        counter += 1
        self.globalIndexSpace_unnecessary = self._global_header_lines[counter]
        counter += 1
        self.timestepsPerLevel_unnecessary = self._global_header_lines[counter]
        counter += 1

        self.dx = np.zeros((self.n_levels, 3))
        for i, line in enumerate(self._global_header_lines[counter:counter + self.n_levels]):
            self.dx[i] = np.array(map(float, line.split()))
        counter += self.n_levels
        self.geometry = int(self._global_header_lines[counter])
        if self.geometry != 0:
            raise RunTimeError("yt only supports cartesian coordinates.")
        counter += 1

        # @todo: this is just to debug. eventually it should go away.
        linebreak = int(self._global_header_lines[counter])
        if linebreak != 0:
            raise RunTimeError("INTERNAL ERROR! This should be a zero.")
        counter += 1

        # each level is one group with ngrids on it. each grid has 3 lines of 2
        # reals
        self.levels = []
        grid_counter = 0
        file_finder_pattern = r"FabOnDisk: (\w+_D_[0-9]{4}) (\d+)\n"
        re_file_finder = re.compile(file_finder_pattern)
        dim_finder_pattern = r"\(\((\d+,\d+,\d+)\) \((\d+,\d+,\d+)\) \(\d+,\d+,\d+\)\)\n"
        re_dim_finder = re.compile(dim_finder_pattern)
        data_files_pattern = r"Level_[\d]/"
        data_files_finder = re.compile(data_files_pattern)

        for level in range(0, self.n_levels):
            tmp = self._global_header_lines[counter].split()
            # should this be grid_time or level_time??
            lev, ngrids, grid_time = int(tmp[0]), int(tmp[1]), float(tmp[2])
            counter += 1
            nsteps = int(self._global_header_lines[counter])
            counter += 1
            self.levels.append(NyxLevel(lev, ngrids))

            # Open level header, extract file names and offsets for each grid
            # read slightly out of order here: at the end of the lo, hi pairs
            # for x, y, z is a *list* of files types in the Level directory.
            # Each type has Header and a number of data files (one per
            # processor)
            tmp_offset = counter + 3 * ngrids
            nfiles = 0
            key_off = 0
            files = {}
            offsets = {}
            while nfiles + tmp_offset < len(self._global_header_lines) \
                  and data_files_finder.match(self._global_header_lines[nfiles + tmp_offset]):
                filen = os.path.join(self.parameter_file.path,
                                     self._global_header_lines[nfiles + tmp_offset].strip())
                # open each "_H" header file, and get the number of
                # components within it
                level_header_file = open(filen + '_H', 'r').read()
                start_stop_index = re_dim_finder.findall(level_header_file) # just take the last one
                grid_file_offset = re_file_finder.findall(level_header_file)
                ncomp_this_file = int(level_header_file.split('\n')[2])

                for i in range(ncomp_this_file):
                    key = self.field_list[i + key_off]
                    f, o = zip(*grid_file_offset)
                    files[key] = f
                    offsets[key] = o
                    self.field_indexes[key] = i

                key_off += ncomp_this_file
                nfiles += 1

            # convert dict of lists to list of dicts
            fn = []
            off = []
            lead_path = os.path.join(self.parameter_file.path,
                                     'Level_%i' % level)
            for i in range(ngrids):
                fi = [os.path.join(lead_path, files[key][i]) for key in self.field_list]
                of = [int(offsets[key][i]) for key in self.field_list]
                fn.append(dict(zip(self.field_list, fi)))
                off.append(dict(zip(self.field_list, of)))

            for grid in range(0, ngrids):
                gfn = fn[grid]  # filename of file containing this grid
                gfo = off[grid] # offset within that file
                xlo, xhi = map(float, self._global_header_lines[counter].split())
                counter += 1
                ylo, yhi = map(float, self._global_header_lines[counter].split())
                counter += 1
                zlo, zhi = map(float, self._global_header_lines[counter].split())
                counter += 1
                lo = np.array([xlo, ylo, zlo])
                hi = np.array([xhi, yhi, zhi])
                dims, start, stop = self.__calculate_grid_dimensions(start_stop_index[grid])
                self.levels[-1].grids.append(self.grid(lo, hi, grid_counter,
                                             level, gfn, gfo, dims, start, stop,
                                             hierarchy=self))
                grid_counter += 1 # this is global, and shouldn't be reset
                                  # for each level

            # already read the filenames above...
            counter += nfiles
            self.num_grids = grid_counter
            self.float_type = 'float64'

        self.maxLevel = self.n_levels - 1
        self.max_level = self.n_levels - 1
        header_file.close()

    def read_particle_header(self):
        # We need to get particle offsets and particle counts
        if not self.parameter_file.use_particles:
            self.pgrid_info = np.zeros((self.num_grids, 3), dtype='int64')
            return
        self.field_list += nyx_particle_field_names[:]
        header = open(os.path.join(self.parameter_file.path, "DM", "Header"))
        version = header.readline()
        ndim = header.readline()
        nfields = header.readline()
        ntotalpart = int(header.readline())
        dummy = header.readline() # nextid
        maxlevel = int(header.readline()) # max level

        # Skip over how many grids on each level; this is degenerate
        for i in range(maxlevel + 1):dummy = header.readline()

        grid_info = np.fromiter((int(i) for line in header.readlines()
                                 for i in line.split()),
                                dtype='int64',
                                count=3*self.num_grids).reshape((self.num_grids, 3))
        # we need grid_info in `populate_grid_objects`, so save it to self
        self.pgrid_info = grid_info

    def __cache_endianness(self, test_grid):
        """
        Cache the endianness and bytes per real of the grids by using a test
        grid and assuming that all grids have the same endianness. This is a
        pretty safe assumption since Nyx uses one file per processor (@todo:
        make sure this is still true, I don't think so). If you are running on a
        cluster with different endian processors, then you are disappoint.

        """
        # open the test file & grab the header
        inFile = open(os.path.expanduser(test_grid.filename[self.field_list[0]]), 'rb')
        header = inFile.readline()
        inFile.close()
        header.strip()

        headerRe = re.compile(fab_header_pattern)
        bytesPerReal, endian, start, stop, centerType, nComponents = \
            headerRe.search(header).groups()
        self._bytesPerReal = int(bytesPerReal)
        if self._bytesPerReal == int(endian[0]):
            dtype = '<'
        elif self._bytesPerReal == int(endian[-1]):
            dtype = '>'
        else:
            raise ValueError("FAB header is neither big nor little endian. Perhaps the file is corrupt?")

        dtype += ('f%i' % self._bytesPerReal) # always a floating point
        self._dtype = dtype

    def __calculate_grid_dimensions(self, start_stop):
        start = np.array(map(int, start_stop[0].split(',')))
        stop = np.array(map(int, start_stop[1].split(',')))
        dimension = stop - start + 1
        return dimension, start, stop

    def _populate_grid_objects(self):
        mylog.debug("Creating grid objects")

        self.grids = np.concatenate([level.grids for level in self.levels])
        basedir = self.parameter_file.path
        for g, pg in itertools.izip(self.grids, self.pgrid_info):
            g.particle_filename = os.path.join(basedir, "DM",
                                               "Level_%s" % (g.Level),
                                               "DATA_%04i" % pg[0])
            g.NumberOfParticles = pg[1]
            g._particle_offset = pg[2]

        self.grid_particle_count[:, 0] = self.pgrid_info[:, 1]
        del self.pgrid_info

        gls = np.concatenate([level.ngrids * [level.level] for level in self.levels])
        self.grid_levels[:] = gls.reshape((self.num_grids, 1))
        grid_dcs = np.concatenate([level.ngrids*[self.dx[level.level]]
                                   for level in self.levels], axis=0)

        self.grid_dxs = grid_dcs[:, 0].reshape((self.num_grids, 1))
        self.grid_dys = grid_dcs[:, 1].reshape((self.num_grids, 1))
        self.grid_dzs = grid_dcs[:, 2].reshape((self.num_grids, 1))

        left_edges = []
        right_edges = []
        dims = []
        for level in self.levels:
            left_edges += [g.LeftEdge for g in level.grids]
            right_edges += [g.RightEdge for g in level.grids]
            dims += [g.ActiveDimensions for g in level.grids]

        self.grid_left_edge = np.array(left_edges)
        self.grid_right_edge = np.array(right_edges)
        self.grid_dimensions = np.array(dims)
        self.gridReverseTree = [] * self.num_grids
        self.gridReverseTree = [ [] for i in range(self.num_grids)]  # why the same thing twice?
        self.gridTree = [ [] for i in range(self.num_grids)]

        mylog.debug("Done creating grid objects")

    def _populate_hierarchy(self):
        self.__setup_grid_tree()

        for i, grid in enumerate(self.grids):
            if (i % 1e4) == 0:
                mylog.debug("Prepared % 7i / % 7i grids", i, self.num_grids)

            grid._prepare_grid()
            grid._setup_dx()

    def __setup_grid_tree(self):
        mask = np.empty(self.grids.size, dtype='int32')
        for i, grid in enumerate(self.grids):
            get_box_grids_level(grid.LeftEdge, grid.RightEdge, grid.Level + 1,
                                self.grid_left_edge, self.grid_right_edge,
                                self.grid_levels, mask)
            children = self.grids[mask.astype("bool")]
            for child in children:
                self.gridReverseTree[child.id].append(i)
                self.gridTree[i].append(weakref.proxy(child))

    def _setup_classes(self):
        dd = self._get_data_reader_dict()
        dd["field_indexes"] = self.field_indexes
        AMRHierarchy._setup_classes(self, dd)
        self.object_types.sort()

    def _get_grid_children(self, grid):
        mask = np.zeros(self.num_grids, dtype='bool')
        grids, grid_ind = self.get_box_grids(grid.LeftEdge, grid.RightEdge)
        mask[grid_ind] = True
        mask = np.logical_and(mask, (self.grid_levels == (grid.Level + 1)).flat)
        return self.grids[mask]

    def _setup_field_list(self):
        if self.parameter_file.use_particles:
            # We know which particle fields will exist -- pending further
            # changes in the future.
            for field in nyx_particle_field_names:
                def external_wrapper(f):
                    def _convert_function(data):
                        return data.convert(f)
                    return _convert_function
                cf = external_wrapper(field)
                # Note that we call add_field on the field_info directly.  This
                # will allow the same field detection mechanism to work for 1D,
                # 2D and 3D fields.
                self.pf.field_info.add_field(field, NullFunc,
                                             convert_function=cf,
                                             take_log=False, particle_type=True)

    def _count_grids(self):
        """ this is already provided in ??? """
        pass

    def _initialize_grid_arrays(self):
        mylog.debug("Allocating arrays for %s grids", self.num_grids)
        self.grid_dimensions = np.ones((self.num_grids, 3), 'int32')
        self.grid_left_edge = np.zeros((self.num_grids, 3), self.float_type)
        self.grid_right_edge = np.ones((self.num_grids, 3), self.float_type)
        self.grid_levels = np.zeros((self.num_grids, 1), 'int32')
        self.grid_particle_count = np.zeros((self.num_grids, 1), 'int32')

    def _parse_hierarchy(self):
        pass

    def _detect_fields(self):
        pass

    def _setup_derived_fields(self):
        self.derived_field_list = []
        for field in self.parameter_file.field_info:
            try:
                fd = self.parameter_file.field_info[field].get_dependencies(
                            pf = self.parameter_file)
            except:
                continue
            available = np.all([f in self.field_list for f in fd.requested])
            if available: self.derived_field_list.append(field)
        for field in self.field_list:
            if field not in self.derived_field_list:
                self.derived_field_list.append(field)

    def _initialize_state_variables(self):
        """
        Override not to re-initialize num_grids in AMRHierarchy.__init__

        """
        self._parallel_locking = False
        self._data_file = None
        self._data_mode = None
        self._max_locations = {}

class NyxLevel:
    def __init__(self, level, ngrids):
        self.level = level
        self.ngrids = ngrids
        self.grids = []

class NyxStaticOutput(StaticOutput):
    """
    This class is a stripped down class that simply reads and parses *filename*,
    without looking at the Nyx hierarchy.

    """
    _hierarchy_class = NyxHierarchy
    _fieldinfo_fallback = NyxFieldInfo
    _fieldinfo_known = KnownNyxFields

    @classmethod
    def _is_valid(cls, *args, **kwargs):
        # fill our args
        pname = args[0].rstrip("/")
        dn = os.path.dirname(pname)
        if len(args) > 1:
            kwargs['paramFilename'] = args[1]

        pfname = kwargs.get("paramFilename", os.path.join(dn, "inputs"))

        # @todo: new Nyx output.
        # We check for the job_info file's existence because this is currently
        # what distinguishes Nyx data from MAESTRO data.
        pfn = os.path.join(pfname)
        if not os.path.exists(pfn): return False
        nyx = any(("nyx." in line for line in open(pfn)))
        maestro = os.path.exists(os.path.join(pname, "job_info"))
        orion = (not nyx) and (not maestro)
        return nyx

    def __init__(self, plotname, param_filename="inputs",
                 fparam_filename="probin", data_style="nyx_native",
                 storage_filename=None):
        """
        Need to override for Nyx file structure, for now.

        The paramfile is usually called "inputs" and there may be a fortran
        inputs file usually called "probin". `plotname` here will be a directory
        name as per BoxLib, data_style will be one of

         * Native
         * IEEE (not implemented in yt)
         * ASCII (not implemented in yt)

        """
        self.storage_filename = storage_filename
        self.parameter_filename = param_filename
        self.fparameter_filename = fparam_filename

        self.path = os.path.abspath(plotname)  # data folder

        # silly inputs and probin file thing (this is on the Nyx todo list)
        self.parameter_file_path = os.path.join(os.path.dirname(self.path),
                                                self.parameter_filename)

        self.fparameter_file_path = os.path.join(os.path.dirname(self.path),
                                                 self.fparameter_filename)

        self.fparameters = {}

        # @todo: quick fix...
        self.use_particles = False

        # @todo: first line
        # runs ``self._parse_parameter_file()``, ``self._set_units()``, and
        # ``self.print_key_parameters()``
        StaticOutput.__init__(self, plotname.rstrip("/"), data_style=data_style)

        # @todo: check all of these and hopefully factor out of the constructor.
        # These should maybe not be hardcoded?
        self.parameters["HydroMethod"] = "nyx"  # always PPM DE
        self.parameters["Time"] = 1.  # default unit is 1...
        self.parameters["DualEnergyFormalism"] = 0  # always off.
        self.parameters["EOSType"] = -1  # default

        # @todo: hopefully delete this after implementing new Nyx output
        if self.fparameters.has_key("mu"):
            self.parameters["mu"] = self.fparameters["mu"]

    def _parse_parameter_file(self):
        """
        Parses the parameter file and establishes the various dictionaries.

        """
        self._parse_header_file()

        if os.path.isfile(self.fparameter_file_path):
            self._parse_fparameter_file()

        # Let's read the file
        self.unique_identifier = int(os.stat(self.parameter_filename)[ST_CTIME])
        lines = open(self.parameter_file_path).readlines()

        for line in lines:
            if line.find("#") >= 1:  # Keep the commented lines...
                line = line[:line.find("#")]
            line = line.strip()
            if len(line) < 2 or line.find("#") == 0: # ...but skip comments
                continue

            try:
                param, val_string = map(strip, line.split("="))
            except ValueError:
                mylog.error("ValueError: '%s'", line)

            vals = val_string.split()

            # @todo: don't do this here...
            if nyx_to_enzo_dict.has_key(param):
                param_name = nyx_to_enzo_dict[param]
                vals = map(parameter_type_dict[param_name], vals)
                if len(vals) == 1:
                    self.parameters[param_name] = vals[0]
                else:
                    # don't know why this is special
                    if param_name == "RefineBy":
                        self.parameters[param_name] = vals[0]
                    else:
                        self.parameters[param_name] = vals

            elif param.startswith("geometry.prob_hi"):
                self.domain_right_edge = np.array([float(i) for i in vals])
            elif param.startswith("geometry.prob_lo"):
                self.domain_left_edge = np.array([float(i) for i in vals])
            elif param.startswith("particles.write_in_plotfile"):
                self.use_particles = boxlib_bool_to_int(vals[0])
            elif param.startswith("nyx.lo_bc"):
                self.periodicity = ensure_tuple([i == 0 for i in vals])

        # aliases we need
        self.parameters["TopGridRank"] = len(self.parameters["TopGridDimensions"])
        self.dimensionality = self.parameters["TopGridRank"]
        self.domain_dimensions = self.parameters["TopGridDimensions"]
        self.refine_by = self.parameters.get("RefineBy", 2)  # 2 is silent default? Makes sense I suppose.

        # Nyx is always cosmological.
        self.cosmological_simulation = 1
        self.omega_lambda = self.parameters["CosmologyOmegaLambdaNow"]
        self.omega_matter = self.parameters["CosmologyOmegaMatterNow"]
        self.hubble_constant = self.parameters["CosmologyHubbleConstantNow"]

        # Read in the `comoving_a` file and parse the value. We should fix this
        # in the new Nyx output format...
        a_file = open(os.path.join(self.path, "comoving_a"))
        a_string = a_file.readline().strip()
        a_file.close()

        # Set the scale factor and redshift
        self.cosmological_scale_factor = float(a_string)
        self.parameters["CosmologyCurrentRedshift"] = 1 / float(a_string) - 1

        # alias
        self.current_redshift = self.parameters["CosmologyCurrentRedshift"]

    def _parse_header_file(self):
        """
        Parses the BoxLib header file to get any parameters stored there.
        Hierarchy information is read out of this file in NyxHierarchy.

        Currently, only Time is read here.

        """
        header_file = open(os.path.join(self.path, "Header"))
        lines = header_file.readlines()  # hopefully this is small
        header_file.close()

        n_fields = int(lines[1])  # this could change
        self.current_time = float(lines[3 + n_fields])  # fragile

    def _parse_fparameter_file(self):
        """
        Parses the fortran parameter file for Nyx. Most of this will be useless,
        but this is where it keeps mu = mass per particle/m_hydrogen. Also it
        contains the cosmological variables.

        """
        # @todo: delete after new Nyx output
        lines = open(self.fparameter_file_path).readlines()
        for line in lines:
            if line.count("=") == 1:
                nyx_param, val_string = map(strip, line.split("="))

                # Check if we care about this param. If so, translate it.
                if nyx_to_enzo_dict.has_key(nyx_param):
                    param = nyx_to_enzo_dict[nyx_param]
                else:
                    continue

                # parse vals string and correct for fortran double syntax
                vals = val_string.split()
                if val_string.count("'") == 0:  # must be floats
                    vals = map(float, [val.replace('D', 'e').replace('d', 'e')
                                       for val in vals])

                # single element or array?
                if len(vals) == 1:
                    self.parameters[param] = vals[0]
                else:
                    self.parameters[param] = vals

    def _set_units(self):
        """
        Generates the conversion to various physical _units based on the
        parameter file.

        """
        self.units = {}
        self.time_units = {}

        if len(self.parameters) == 0:  # don't think this is possible, but safe
            self._parse_parameter_file()

        # Masses are always in $ M_{\odot} $
        self.units["particle_mass"] = 1.989e33

        mylog.warning("Length units: setting 1.0 = 1.0 Mpc.")
        self.units.update(mpc_conversion)
        self.units["density"] = self.units["particle_mass"]/(self.units["cm"])**3
        self.units["particle_mass_density"] = self.units["density"]
        self.units["Density"] = 1

        # @todo: enzo-isms
        mylog.warning("Time units: setting 1.0 = Mpc/km s ~ 10^12 yr .")
        self.time_units["s"] = 1.0 / 3.08568025e19
        self.conversion_factors["Time"] = 1.0 / 3.08568025e19

        # velocities are not comoving!
        # Nyx is in km/s so we need to convert to cm/s, hence 1e5
        cf = 1e5 * (self.cosmological_scale_factor)
        for ax in "xyz":
            self.units["particle_velocity_%s" % ax] = cf

        # misc
        self.conversion_factors = defaultdict(lambda: 1.0)  # what is this for? - Steffen: this is to get 1.0 for values not in the dict
        self.time_units["1"] = 1
        self.units["1"] = 1.0
        self.units["unitary"] = 1.0 / (self.domain_right_edge -
                                       self.domain_left_edge).max()

        # time
        for unit in sec_conversion.keys():
            self.time_units[unit] = self.time_units["s"] / sec_conversion[unit]

        # not the most useful right now, but someday
        for key in nyx_particle_field_names:
            self.conversion_factors[key] = 1.0

    def _setup_nounits_units(self):
        z = 0

    def _localize(self, f, default):
        if f is None:
            return os.path.join(self.directory, default)
        return f
